# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module contains useful methods for handling xml trees and files which are used
by the Fleur code and the fleur plugin.
"""
# TODO FEHLER meldungen, currently if a xpath expression is valid, but does not exists
# xpath returns []. Do we want this behavior?
# TODO finish implementation of create=False
# TODO: no aiida imports

from __future__ import absolute_import
from __future__ import print_function
from lxml import etree
import six
from six.moves import range

from aiida.common.exceptions import InputValidationError


def is_sequence(arg):
    """
    Checks if arg is a sequence
    """
    if isinstance(arg, str):
        return False
    return (not hasattr(arg, "strip") and
            hasattr(arg, "__getitem__") or
            hasattr(arg, "__iter__"))

##### CONVERTERS ############

def convert_to_float(value_string, parser_info_out=None, suc_return=True):
    """
    Tries to make a float out of a string. If it can't it logs a warning
    and returns True or False if convertion worked or not.

    :param value_string: a string
    :returns value: the new float or value_string: the string given
    :returns: True or False
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}
    try:
        value = float(value_string)
    except TypeError:
        parser_info_out['parser_warnings'].append('Could not convert: "{}" to float, TypeError'
                                                  ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    except ValueError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to float, ValueError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    if suc_return:
        return value, True
    else:
        return value


def convert_to_int(value_string, parser_info_out=None, suc_return=True):
    """
    Tries to make a int out of a string. If it can't it logs a warning
    and returns True or False if convertion worked or not.

    :param value_string: a string
    :returns value: the new int or value_string: the string given
    :returns: True or False, if suc_return=True
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}
    try:
        value = int(value_string)
    except TypeError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to int, TypeError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    except ValueError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to int, ValueError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    if suc_return:
        return value, True
    else:
        return value


def convert_htr_to_ev(value, parser_info_out=None):
    """
    Multiplies the value given with the Hartree factor (converts htr to eV)
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}
    htr = 27.21138602
    suc = False
    value_to_save, suc = convert_to_float(value, parser_info_out=parser_info_out)
    if suc:
        return value_to_save * htr
    else:
        return value


def convert_ev_to_htr(value, parser_info_out=None):
    """
    Divides the value given with the Hartree factor (converts htr to eV)
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}
    htr = 27.21138602
    suc = False
    value_to_save, suc = convert_to_float(value, parser_info_out=parser_info_out)
    if suc:
        return value_to_save / htr
    else:
        return value


def convert_from_fortran_bool(stringbool):
    """
    Converts a string in this case ('T', 'F', or 't', 'f') to True or False

    :param stringbool: a string ('t', 'f', 'F', 'T')

    :return: boolean  (either True or False)
    """
    true_items = ['True', 't', 'T']
    false_items = ['False', 'f', 'F']
    if isinstance(stringbool, str):
        if stringbool in false_items:
            return False
        if stringbool in true_items:
            return True
        else:
            raise InputValidationError(
                "A string: {} for a boolean was given, which is not 'True',"
                " 'False', 't', 'T', 'F' or 'f'".format(stringbool))
    elif isinstance(stringbool, bool):
        return stringbool  # no convertion needed...
    else:
        raise TypeError("convert_to_fortran_bool accepts only a string or "
                        "bool as argument")


def convert_to_fortran_bool(boolean):
    """
    Converts a Boolean as string to the format defined in the input

    :param boolean: either a boolean or a string ('True', 'False', 'F', 'T')

    :return: a string (either 't' or 'f')
    """

    if isinstance(boolean, bool):
        if boolean:
            new_string = 'T'
            return new_string
        else:
            new_string = 'F'
            return new_string
    elif isinstance(boolean, str):  # basestring):
        if boolean == 'True' or boolean == 't' or boolean == 'T':
            new_string = 'T'
            return new_string
        elif boolean == 'False' or boolean == 'f' or boolean == 'F':
            new_string = 'F'
            return new_string
        else:
            raise InputValidationError(
                "A string: {} for a boolean was given, which is not 'True',"
                "'False', 't', 'T', 'F' or 'f'".format(boolean))
    else:
        raise TypeError("convert_to_fortran_bool accepts only a string or "
                        "bool as argument, given {} ".format(boolean))


def convert_to_fortran_string(string):
    """
    converts some parameter strings to the format for the inpgen
    :param string: some string
    :returns: string in right format (extra "")
    """
    new_string = '"' + string + '"'
    return new_string


def convert_fleur_lo(loelements):
    """
    Converts lo xml elements from the inp.xml file into a lo string for the inpgen
    """
    # Developer hint: Be careful with using '' and "", basestring and str are not the same...
    # therefore other conversion methods might fail, or the wrong format could be written.
    from aiida_fleur.tools.element_econfig_list import shell_map

    lo_string = ''
    for element in loelements:
        lo_type = get_xml_attribute(element, 'type')
        if lo_type != 'SCLO':  # non standard los not supported for now
            continue
        l_num = get_xml_attribute(element, 'l')
        n_num = get_xml_attribute(element, 'n')
        l_char = shell_map.get(int(l_num), '')
        lostr = '{}{}'.format(n_num, l_char)
        lo_string = lo_string + ' ' + lostr
    return lo_string.strip()


def set_dict_or_not(para_dict, key, value):
    """
    setter method for a dictionary that will not set the key, value pair.
    if the key is [] or None.
    """
    if value == [] or value is None:
        return para_dict
    else:
        para_dict[key] = value
        return para_dict

####### XML SETTERS GENERAL ##############


def xml_set_attribv_occ(xmltree, xpathn, attributename, attribv, occ=None, create=False):
    """
    Routine sets the value of an attribute in the xml file on only the places
    specified in occ

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute
    :param attributename: an attribute name
    :param attribv: an attribute value which will be set
    :param occ: a list of integers specifying number of occurrence to be set
    :param create: if True and there is no given xpath in the FleurinpData, creates it

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'band', 'T')
             xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'dos', 'F')
    """
    if occ is None:
        occ = [0]

    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpathn, create=create)

    if not isinstance(attribv, type('')):
        attribv = str(attribv)
    for i, node in enumerate(nodes):
        if i in occ:
            node.set(attributename, attribv)
        if -1 in occ:  # 'all'
            node.set(attributename, attribv)


def xml_set_first_attribv(xmltree, xpathn, attributename, attribv, create=False):
    """
    Routine sets the value of the first found attribute in the xml file

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute
    :param attributename: an attribute name
    :param attribv: an attribute value which will be set
    :param create: if True and there is no given xpath in the FleurinpData, creates it

    :return: None, or an etree

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'band', 'T')
             xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'dos', 'F')
    """

    root = xmltree.getroot()
    if isinstance(attribv, type('')):
        eval_xpath3(root, xpathn, create=create)[0].set(attributename, attribv)
    else:
        eval_xpath3(root, xpathn, create=create)[0].set(attributename, str(attribv))
    # return xmltree
    # ToDO check if worked. else exception,


def xml_set_all_attribv(xmltree, xpathn, attributename, attribv, create=False):
    """
    Routine sets the value of an attribute in the xml file on all places it occurs

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute
    :param attributename: an attribute name
    :param attribv: an attribute value which will be set
    :param create: if True and there is no given xpath in the FleurinpData, creates it

    :return: None, or an etree

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: xml_set_first_attribv(tree, '/fleurInput/atomGroups/atomGroup/force', 'relaxXYZ', 'TTF')
             xml_set_first_attribv(tree, '/fleurInput/atomGroups/atomGroup/force', 'calculate', 'F')
    """

    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpathn, create=create)
    if is_sequence(attribv):
        for i, node in enumerate(nodes):
            if not isinstance(attribv[i], str):  # type(attribv) != type(''):
                attribv[i] = str(attribv[i])
            node.set(attributename, attribv[i])
    else:
        if not isinstance(attribv, str):  # type(attribv) != type(''):
            attribv = str(attribv)
        for node in nodes:
            node.set(attributename, attribv)


def xml_set_text(xmltree, xpathn, text, create=False, place_index=None, tag_order=None):
    """
    Routine sets the text of a tag in the xml file

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute
    :param text: text to be set
    :param create: if True and there is no given xpath in the FleurinpData, creates it
    :param place_index: if create=True, defines the place where to put a created tag
    :param tag_order: if create=True, defines a tag order

    example:

        xml_set_text(tree, '/fleurInput/comment', 'Test Fleur calculation for AiiDA plug-in')

    but also coordinates and Bravais Matrix!:

        xml_set_text(tree, '/fleurInput/atomGroups/atomGroup/relPos','1.20000 PI/3 5.1-MYCrazyCostant')
    """

    root = xmltree.getroot()
    node = eval_xpath3(root, xpathn, create=create, place_index=place_index, tag_order=tag_order)
    if node:
        node[0].text = text
    # return xmltree


def xml_set_text_occ(xmltree, xpathn, text, create=False, occ=0, place_index=None, tag_order=None):
    """
    Routine sets the text of a tag in the xml file

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute
    :param text: text to be set
    :param create: if True and there is no given xpath in the FleurinpData, creates it
    :param occ: an integer that sets occurrence number to be set
    :param place_index: if create=True, defines the place where to put a created tag
    :param tag_order: if create=True, defines a tag order
    """

    root = xmltree.getroot()
    node = eval_xpath3(root, xpathn, create=create, place_index=place_index, tag_order=tag_order)
    if node:
        node[occ].text = text


def xml_set_all_text(xmltree, xpathn, text, create=False, tag_order=None):
    """
    Routine sets the text of a tag in the xml file

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute
    :param text: text to be set
    :param create: if True and there is no given xpath in the FleurinpData, creates it
    :param place_index: if create=True, defines the place where to put a created tag
    :param tag_order: if create=True, defines a tag order
    """
    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpathn, create=create, tag_order=tag_order)
    if is_sequence(text):
        for i, node in enumerate(nodes):
            node.text = text[i]
    else:
        for node in nodes:
            node.text = text


def create_tag(xmlnode, xpath, newelement, create=False, place_index=None, tag_order=None):
    """
    This method evaluates an xpath expresion and creates tag in an xmltree under the
    returned nodes. If the path does exist things will be overwritten, or created.
    Per default the new element is appended to the elements, but it can also be
    inserted in a certain position or after certain other tags.

    :param xmlnode: an xmltree that represents inp.xml
    :param xpathn: a path where to place a new tag
    :param newelement: a tag name to be created
    :param create: if True and there is no given xpath in the FleurinpData, creates it
    :param place_index: defines the place where to put a created tag
    :param tag_order: defines a tag order
    """
    import copy
    newelement_name = newelement
    if not etree.iselement(newelement):
        try:
            newelement = etree.Element(newelement)
        except ValueError as v:
            raise ValueError('{}. If this is a species, are you sure this species exists '
                             'in your inp.xml?'.format(v))
    nodes = eval_xpath3(xmlnode, xpath, create=create)
    if nodes:
        for node_1 in nodes:
            element_to_write = copy.deepcopy(newelement)
            if place_index:
                if tag_order:
                    # behind what shall I place it
                    try:
                        place_index = tag_order.index(newelement_name)
                    except:
                        raise ValueError('Did not find element name in the tag_order list')
                    behind_tags = tag_order[:place_index]
                    # check if children are in the same sequence as given in tag_order
                    tags = []
                    for child in node_1.iterchildren():
                        if child.tag not in tags:
                            tags.append(child.tag)
                    prev = -1
                    for name in tags:
                        try:
                            current = tag_order.index(name)
                        except ValueError:
                            raise ValueError('Did not find existing tag name in the tag_order list'
                                             ': {}'.format(name))
                        if current > prev:
                            prev = current
                        else:
                            raise ValueError('Existing order does not correspond to tag_order list')
                    # get all names of tag existing tags
                    was_set = False
                    for tag in reversed(behind_tags):
                        for child in node_1.iterchildren(tag=tag, reversed=False):
                            # if tagname of elements==tag:
                            tag_index = node_1.index(child)
                            try:
                                node_1.insert(tag_index + 1, element_to_write)
                            except ValueError as v:
                                raise ValueError('{}. If this is a species, are'
                                                 'you sure this species exists in your inp.xml?'
                                                 ''.format(v))
                            was_set = True
                            break
                        if was_set:
                            break
                    if not was_set:  # just append
                        try:
                            node_1.insert(0, element_to_write)
                        except ValueError as v:
                            raise ValueError('{}. If this is a species, are you'
                                             ' sure this species exists in your inp.xml?'
                                             ''.format(v))
                    # (or remove all and write them again in right order?)
                else:
                    try:
                        node_1.insert(place_index, element_to_write)
                    except ValueError as v:
                        raise ValueError('{}. If this is a species, are you sure this species '
                                         'exists in your inp.xml?'.format(v))
            else:
                try:
                    node_1.append(element_to_write)
                except ValueError as v:
                    raise ValueError('{}. If this is a species, are you sure this species exists'
                                     'in your inp.xml?'.format(v))
    return xmlnode


def delete_att(xmltree, xpath, attrib):
    """
    Deletes an xml tag in an xmletree.

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the attribute to be deleted
    :param attrib: the name of an attribute
    """
    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpath)
    if nodes:
        for node in nodes:
            try:
                del node.attrib[attrib]
            except BaseException:
                pass
    return xmltree


def delete_tag(xmltree, xpath):
    """
    Deletes an xml tag in an xmletree.

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the tag to be deleted
    """
    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpath)
    if nodes:
        for node in nodes:
            parent = node.getparent()
            parent.remove(node)
    return xmltree


def replace_tag(xmltree, xpath, newelement):
    """
    replaces a xml tag by another tag on an xmletree in place

    :param xmltree: an xmltree that represents inp.xml
    :param xpathn: a path to the tag to be replaced
    :param newelement: a new tag
    """
    root = xmltree.getroot()

    nodes = eval_xpath3(root, xpath)
    if nodes:
        for node in nodes:
            parent = node.getparent()
            index = parent.index(node)
            parent.remove(node)
            parent.insert(index, newelement)

    return xmltree


def get_inpgen_paranode_from_xml(inpxmlfile):
    """
    This routine returns an AiiDA Parameter Data type produced from the inp.xml
    file, which can be used by inpgen.

    :return: ParameterData node
    """
    from aiida.orm import Dict
    para_dict = get_inpgen_para_from_xml(inpxmlfile)
    return Dict(dict=para_dict)


def get_inpgen_para_from_xml(inpxmlfile):
    """
    This routine returns an python dictionary produced from the inp.xml
    file, which can be used as a calc_parameters node by inpgen.
    Be aware that inpgen does not take all information that is contained in an inp.xml file

    :return new_parameters: A Dict, which will lead to the same inp.xml (in case if other defaults,
                            which can not be controlled by input for inpgen, were changed)

    """

    # TODO: convert econfig
    # TODO: parse kpoints, somehow count is bad (if symmetry changes), mesh is not known, path cannot be specified

    # Disclaimer: this routine needs some xpath expressions. these are hardcoded here,
    # therefore maintainance might be needed, if you want to circumvent this, you have
    # to get all the paths from somewhere.

    #######
    # all hardcoded xpaths used and attributes names:
    # input
    film_xpath = '/fleurInput/atomGroups/atomGroup/filmPos/'  # check for film pos

    # atom, for each species\
    species_xpath = '/fleurInput/atomSpecies/species'
    atom_id_xpath = ''  # is reconstruction possible at all now?
    atom_z_xpath = '@atomicNumber'
    atom_rmt_xpath = 'mtSphere/@radius'
    atom_dx_xpath = 'mtSphere/@logIncrement'
    atom_jri_xpath = 'mtSphere/@gridPoints'
    atom_lmax_xpath = 'atomicCutoffs/@lmax'
    atom_lnosph_xpath = 'atomicCutoffs/@lnonsphr'
    atom_ncst_xpath = '@coreStates'
    atom_econfig_xpath = 'electronConfig'  # converting todo
    atom_bmu_xpath = '@magMom'
    atom_lo_xpath = 'lo'  # converting todo
    atom_element_xpath = '@element'
    atom_name_xpath = '@name'

    # comp
    jspins_xpath = 'calculationSetup/magnetism/@jspins'
    frcor_xpath = 'calculationSetup/coreElectrons/@frcor'
    ctail_xpath = 'calculationSetup/coreElectrons/@ctail'
    kcrel_xpath = 'calculationSetup/coreElectrons/@kcrel'
    gmax_xpath = 'calculationSetup/cutoffs/@Gmax'
    gmaxxc_xpath = 'calculationSetup/cutoffs/@GmaxXC'
    kmax_xpath = 'calculationSetup/cutoffs/@Kmax'

    # exco
    exco_xpath = 'xcFunctional/@name'
    # film

    # soc
    l_soc_xpath = '//calculationSetup/soc/@l_soc'
    theta_xpath = '//calculationSetup/soc/@theta'
    phi_xpath = '//calculationSetup/soc/@phi'
    # qss

    # kpt

    title_xpath = '/fleurInput/comment/text()'  # text

    ########
    new_parameters = {}

    #print('parsing inp.xml without XMLSchema')
    #tree = etree.parse(inpxmlfile)
    tree = inpxmlfile
    root = tree.getroot()

    # Create the cards

    # &input # most things are not needed for AiiDA here. or we ignor them for now.
    # film is set by the plugin depended on the structure
    # symor per default = False? to avoid input which fleur can't take

    # &comp
    # attrib = get_xml_attribute(
    comp_dict = {}
    comp_dict = set_dict_or_not(comp_dict, 'jspins',
                                convert_to_float(eval_xpath(root, jspins_xpath), suc_return=False))
    comp_dict = set_dict_or_not(comp_dict, 'frcor',
                                convert_from_fortran_bool(eval_xpath(root, frcor_xpath)))
    comp_dict = set_dict_or_not(comp_dict, 'ctail',
                                convert_from_fortran_bool(eval_xpath(root, ctail_xpath)))
    comp_dict = set_dict_or_not(comp_dict, 'kcrel', eval_xpath(root, kcrel_xpath))
    comp_dict = set_dict_or_not(comp_dict, 'gmax',
                                convert_to_float(eval_xpath(root, gmax_xpath), suc_return=False))
    comp_dict = set_dict_or_not(comp_dict, 'gmaxxc',
                                convert_to_float(eval_xpath(root, gmaxxc_xpath), suc_return=False))
    comp_dict = set_dict_or_not(comp_dict, 'kmax',
                                convert_to_float(eval_xpath(root, kmax_xpath), suc_return=False))
    new_parameters['comp'] = comp_dict

    # &atoms
    species_list = eval_xpath2(root, species_xpath)

    for i, species in enumerate(species_list):
        atom_dict = {}
        atoms_name = 'atom{}'.format(i)
        atom_z = convert_to_int(eval_xpath(species, atom_z_xpath), suc_return=False)
        atom_rmt = convert_to_float(eval_xpath(species, atom_rmt_xpath), suc_return=False)
        atom_dx = convert_to_float(eval_xpath(species, atom_dx_xpath), suc_return=False)
        atom_jri = convert_to_int(eval_xpath(species, atom_jri_xpath), suc_return=False)
        atom_lmax = convert_to_int(eval_xpath(species, atom_lmax_xpath), suc_return=False)
        atom_lnosph = convert_to_int(eval_xpath(species, atom_lnosph_xpath), suc_return=False)
        atom_ncst = convert_to_int(eval_xpath(species, atom_ncst_xpath), suc_return=False)
        atom_econfig = eval_xpath(species, atom_econfig_xpath)
        atom_bmu = convert_to_float(eval_xpath(species, atom_bmu_xpath), suc_return=False)
        atom_lo = eval_xpath(species, atom_lo_xpath)
        atom_element = eval_xpath(species, atom_element_xpath)
        atom_name_2 = eval_xpath(species, atom_name_xpath)

        atom_dict = set_dict_or_not(atom_dict, 'z', atom_z)
        atom_dict = set_dict_or_not(atom_dict, 'rmt', atom_rmt)
        atom_dict = set_dict_or_not(atom_dict, 'dx', atom_dx)
        atom_dict = set_dict_or_not(atom_dict, 'jri', atom_jri)
        atom_dict = set_dict_or_not(atom_dict, 'lmax', atom_lmax)
        atom_dict = set_dict_or_not(atom_dict, 'lnonsph', atom_lnosph)
        atom_dict = set_dict_or_not(atom_dict, 'ncst', atom_ncst)
        atom_dict = set_dict_or_not(atom_dict, 'econfig', atom_econfig)
        atom_dict = set_dict_or_not(atom_dict, 'bmu', atom_bmu)
        if atom_lo is not None:
            atom_dict = set_dict_or_not(atom_dict, 'lo', convert_fleur_lo(atom_lo))
        atom_dict = set_dict_or_not(atom_dict, 'element', '{}'.format(atom_element))
        #atom_dict = set_dict_or_not(atom_dict, 'name', atom_name_2)

        new_parameters[atoms_name] = atom_dict

    # &soc
    attrib = convert_from_fortran_bool(eval_xpath(root, l_soc_xpath))
    theta = convert_to_float(eval_xpath(root, theta_xpath), suc_return=False)
    phi = convert_to_float(eval_xpath(root, phi_xpath), suc_return=False)
    if attrib:
        new_parameters['soc'] = {'theta': theta, 'phi': phi}

    # &kpt
    #attrib = convert_from_fortran_bool(eval_xpath(root, l_soc_xpath))
    #theta = eval_xpath(root, theta_xpath)
    #phi = eval_xpath(root, phi_xpath)
    # if kpt:
    #    new_parameters['kpt'] = {'theta' : theta, 'phi' : phi}
    #    # ['nkpt', 'kpts', 'div1', 'div2', 'div3',                         'tkb', 'tria'],

    # title
    title = eval_xpath(root, title_xpath)  # text
    if title:
        new_parameters['title'] = title.replace('\n', '').strip()

    # &exco
    #TODO, easy
    exco_dict = {}
    exco_dict = set_dict_or_not(exco_dict, 'xctyp', eval_xpath(root, exco_xpath))
    # 'exco' : ['xctyp', 'relxc'],
    new_parameters['exco'] = exco_dict
    # &film
    # TODO

    # &qss
    # TODO

    # lattice, not supported?

    return new_parameters

####### XML SETTERS SPECIAL ########

def set_species_label(fleurinp_tree_copy, at_label, attributedict, create=False):
    """
    This method calls :func:`~aiida_fleur.tools.xml_util.set_species()`
    method for a certain atom specie that corresponds to an atom with a given label

    :param fleurinp_tree_copy: xml etree of the inp.xml
    :param at_label: string, a label of the atom which specie will be changed
    :param attributedict: a python dict specifying what you want to change.
    :param create: bool, if species does not exist create it and all subtags?
    """

    if at_label == 'all':
        fleurinp_tree_copy = set_species(fleurinp_tree_copy, 'all', attributedict, create)
        return fleurinp_tree_copy

    specie = ''
    at_label = "{: >20}".format(at_label)
    all_groups = eval_xpath2(fleurinp_tree_copy, '/fleurInput/atomGroups/atomGroup')

    for group in all_groups:
        positions = eval_xpath2(group, 'filmPos')
        if not positions:
            positions = eval_xpath2(group, 'relPos')
        for atom in positions:
            atom_label = get_xml_attribute(atom, 'label')
            if atom_label == at_label:
                specie = get_xml_attribute(group, 'species')

    fleurinp_tree_copy = set_species(fleurinp_tree_copy, specie, attributedict, create)

    return fleurinp_tree_copy


def set_species(fleurinp_tree_copy, species_name, attributedict, create=False):
    """
    Method to set parameters of a species tag of the fleur inp.xml file.

    :param fleurinp_tree_copy: xml etree of the inp.xml
    :param species_name: string, name of the specie you want to change
    :param attributedict: a python dict specifying what you want to change.
    :param create: bool, if species does not exist create it and all subtags?

    :raises ValueError: if species name is non existent in inp.xml and should not be created.
                        also if other given tags are garbage. (errors from eval_xpath() methods)

    :return fleurinp_tree_copy: xml etree of the new inp.xml

    **attributedict** is a python dictionary containing dictionaries that specify attributes
    to be set inside the certain specie. For example, if one wants to set a MT radius it
    can be done via::

        attributedict = {'mtSphere' : {'radius' : 2.2}}

    Another example::

        'attributedict': {'special': {'socscale': 0.0}}

    that switches SOC terms on a sertain specie. ``mtSphere``, ``atomicCutoffs``,
    ``energyParameters``, ``lo``, ``electronConfig``, ``nocoParams``, ``ldaU`` and
    ``special`` keys are supported. To find possible
    keys of the inner dictionary please refer to the FLEUR documentation flapw.de
    """
    # TODO lowercase everything
    # TODO make a general specifier for species, not only the name i.e. also
    # number, other parameters
    if species_name == 'all':
        xpath_species = '/fleurInput/atomSpecies/species'
    else:
        xpath_species = '/fleurInput/atomSpecies/species[@name = "{}"]'.format(species_name)

    xpath_mt = '{}/mtSphere'.format(xpath_species)
    xpath_atomic_cutoffs = '{}/atomicCutoffs'.format(xpath_species)
    xpath_energy_parameters = '{}/energyParameters'.format(xpath_species)
    xpath_lo = '{}/lo'.format(xpath_species)
    xpath_electron_config = '{}/electronConfig'.format(xpath_species)
    xpath_core_occ = '{}/electronConfig/stateOccupation'.format(xpath_species)
    xpath_lda_u = '{}/ldaU'.format(xpath_species)
    xpath_soc_scale = '{}/special'.format(xpath_species)

    # can we get this out of schema file?
    species_seq = [
        'mtSphere',
        'atomicCutoffs',
        'energyParameters',
        'prodBasis',
        'special',
        'force',
        'electronConfig',
        'nocoParams',
        'ldaU',
        'lo']

    for key, val in six.iteritems(attributedict):
        if key == 'mtSphere':  # always in inp.xml
            for attrib, value in six.iteritems(val):
                xml_set_all_attribv(fleurinp_tree_copy, xpath_mt, attrib, value)
        elif key == 'atomicCutoffs':  # always in inp.xml
            for attrib, value in six.iteritems(val):
                xml_set_all_attribv(fleurinp_tree_copy, xpath_atomic_cutoffs, attrib, value)
        elif key == 'energyParameters':  # always in inp.xml
            for attrib, value in six.iteritems(val):
                xml_set_all_attribv(fleurinp_tree_copy, xpath_energy_parameters, attrib, value)
        elif key == 'lo':  # optional in inp.xml
            # policy: we DELETE all LOs, and create new ones from the given parameters.
            existinglos = eval_xpath3(fleurinp_tree_copy, xpath_lo)
            for los in existinglos:
                parent = los.getparent()
                parent.remove(los)

            # there can be multible LO tags, so I expect either one or a list
            if isinstance(val, dict):
                create_tag(
                    fleurinp_tree_copy,
                    xpath_species,
                    'lo',
                    place_index=species_seq.index('lo'),
                    tag_order=species_seq)
                for attrib, value in six.iteritems(val):
                    xml_set_all_attribv(fleurinp_tree_copy, xpath_lo, attrib, value, create=True)
            else:  # I expect a list of dicts
                # lonodes = eval_xpath3(root, xpathlo)#, create=True, place_index=species_seq.index('lo'), tag_order=species_seq)
                #nlonodes = len(lonodes)
                # ggf create more lo tags of needed
                los_need = len(val)  # - nlonodes
                for j in range(0, los_need):
                    create_tag(
                        fleurinp_tree_copy,
                        xpath_species,
                        'lo',
                        place_index=species_seq.index('lo'),
                        tag_order=species_seq)
                for i, lodict in enumerate(val):
                    for attrib, value in six.iteritems(lodict):
                        sets = []
                        for k in range(len(eval_xpath2(fleurinp_tree_copy, xpath_species + '/lo'))//los_need):
                            sets.append(k * los_need + i)
                        xml_set_attribv_occ(fleurinp_tree_copy, xpath_lo, attrib, value, occ=sets)

        elif key == 'electronConfig':
            # eval electronConfig and ggf create tag at right place.
            eval_xpath3(
                fleurinp_tree_copy,
                xpath_electron_config,
                create=True,
                place_index=species_seq.index('electronConfig'),
                tag_order=species_seq)

            for tag in ['coreConfig', 'valenceConfig', 'stateOccupation']:
                for etag, edictlist in six.iteritems(val):
                    if not etag == tag:
                        continue
                    if etag == 'stateOccupation':  # there can be multiple times stateOccupation
                        # policy: default we DELETE all existing occs and create new ones for the
                        # given input!
                        existingocc = eval_xpath3(fleurinp_tree_copy, xpath_core_occ)
                        for occ in existingocc:
                            parent = occ.getparent()
                            parent.remove(occ)
                        if isinstance(edictlist, dict):
                            for attrib, value in six.iteritems(edictlist):
                                xml_set_all_attribv(
                                    fleurinp_tree_copy, xpath_core_occ, attrib, value, create=True)
                        else:  # I expect a list of dicts
                            nodes_need = len(edictlist)
                            for j in range(0, nodes_need):
                                create_tag(
                                    fleurinp_tree_copy,
                                    xpath_electron_config,
                                    'stateOccupation',
                                    create=True)
                            for i, occdict in enumerate(edictlist):
                                # override them one after one
                                sets = []
                                for k in range(len(eval_xpath2(fleurinp_tree_copy, xpath_core_occ))//nodes_need):
                                    sets.append(k * nodes_need + i)
                                for attrib, value in six.iteritems(occdict):
                                    xml_set_attribv_occ(
                                        fleurinp_tree_copy, xpath_core_occ, attrib, value, occ=sets)

                    else:
                        xpathconfig = xpath_electron_config + '/{}'.format(etag)
                        xml_set_all_text(
                            fleurinp_tree_copy,
                            xpathconfig,
                            edictlist,
                            create=create,
                            tag_order=['coreConfig', 'valenceConfig', 'stateOccupation'])
        elif key == 'ldaU':
            for attrib, value in six.iteritems(val):
                xml_set_all_attribv(fleurinp_tree_copy, xpath_lda_u, attrib, value, create=True)
        elif key == 'special':
            eval_xpath3(fleurinp_tree_copy,
                        xpath_soc_scale,
                        create=True,
                        place_index=species_seq.index('special'),
                        tag_order=species_seq)
            for attrib, value in six.iteritems(val):
                xml_set_all_attribv(
                    fleurinp_tree_copy,
                    xpath_soc_scale,
                    attrib,
                    value,
                    create=create)
        else:
            xml_set_all_attribv(fleurinp_tree_copy, xpath_species, attrib, value)

    return fleurinp_tree_copy


def shift_value_species_label(fleurinp_tree_copy, at_label, attr_name, value_given, mode='abs'):
    """
    Shifts value of a specie by label
    """
    specie = ''
    at_label = "{: >20}".format(at_label)
    all_groups = eval_xpath2(fleurinp_tree_copy, '/fleurInput/atomGroups/atomGroup')

    for group in all_groups:
        positions = eval_xpath2(group, 'filmPos')
        if not positions:
            positions = eval_xpath2(group, 'relPos')
        for atom in positions:
            atom_label = get_xml_attribute(atom, 'label')
            if atom_label == at_label:
                specie = get_xml_attribute(group, 'species')

    xpath_species = '/fleurInput/atomSpecies/species[@name = "{}"]'.format(specie)
    xpath_mt = '{}/mtSphere'.format(xpath_species)
    xpath_atomic_cutoffs = '{}/atomicCutoffs'.format(xpath_species)
    xpath_energy_parameters = '{}/energyParameters'.format(xpath_species)
    xpath_final = 'initialise'

    if attr_name in ['radius', 'gridPoints', 'logIncrement']:
        xpath_final = xpath_mt
    elif attr_name in ['lmax', 'lnonsphr']:
        xpath_final = xpath_atomic_cutoffs
    elif attr_name in ['s', 'p', 'd', 'f']:
        xpath_final = xpath_energy_parameters

    old_val = eval_xpath2(fleurinp_tree_copy, '/@'.join([xpath_final, attr_name]))

    if not old_val:
        print('Can not find {} attribute in the inp.xml, skip it'.format(attr_name))
    else:
        old_val = float(old_val[0])

    if mode == 'rel':
        value = value_given * old_val
    elif mode == 'abs':
        value = value_given + old_val
    else:
        raise ValueError("Mode should be 'res' or 'abs' only")

    if attr_name in ['radius', 'logIncrement']:
        value_to_write = value
    else:
        if not value.is_integer():
            raise ValueError('You are trying to write a float to an integer attribute')
        value_to_write = int(value)

    xml_set_first_attribv(fleurinp_tree_copy, xpath_final, attr_name,value_to_write)

    return fleurinp_tree_copy


def change_atomgr_att_label(fleurinp_tree_copy, attributedict, at_label):
    """
    This method calls :func:`~aiida_fleur.tools.xml_util.change_atomgr_att()`
    method for a certain atom specie that corresponds to an atom with a given label.
    """

    if at_label == 'all':
        fleurinp_tree_copy = change_atomgr_att(fleurinp_tree_copy, attributedict, position=None,
                                               species='all')
        return fleurinp_tree_copy

    specie = ''
    at_label = "{: >20}".format(at_label)
    all_groups = eval_xpath2(fleurinp_tree_copy, '/fleurInput/atomGroups/atomGroup')

    for group in all_groups:
        positions = eval_xpath2(group, 'filmPos')
        if not positions:
            positions = eval_xpath2(group, 'relPos')
        for atom in positions:
            atom_label = get_xml_attribute(atom, 'label')
            if atom_label == at_label:
                specie = get_xml_attribute(group, 'species')

    fleurinp_tree_copy = change_atomgr_att(fleurinp_tree_copy, attributedict, position=None,
                                           species=specie)

    return fleurinp_tree_copy

def change_atomgr_att(fleurinp_tree_copy, attributedict, position=None, species=None):
    """
    Method to set parameters of an atom group of the fleur inp.xml file.

    :param fleurinp_tree_copy: xml etree of the inp.xml
    :param attributedict: a python dict specifying what you want to change.
    :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
    :param species: atom groups, corresponding to the given specie will be changed
    :param create: bool, if species does not exist create it and all subtags?

    :return fleurinp_tree_copy: xml etree of the new inp.xml

    **attributedict** is a python dictionary containing dictionaries that specify attributes
    to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
    can be done via::

        'attributedict': {'nocoParams': [('beta', val)]}

    ``force`` and ``nocoParams`` keys are supported.
    To find possible keys of the inner dictionary please refer to the FLEUR documentation flapw.de
    """
    xpathatmgroup = '/fleurInput/atomGroups/atomGroup'
    xpathforce = '{}/force'.format(xpathatmgroup)
    xpathnocoParams = '{}/nocoParams'.format(xpathatmgroup)

    if not position and not species:  # not specfied what to change
        return fleurinp_tree_copy

    if position:
        if not position == 'all':
            xpathatmgroup = '/fleurInput/atomGroups/atomGroup[{}]'.format(position)
            xpathforce = '{}/force'.format(xpathatmgroup)
            xpathnocoParams = '{}/nocoParams'.format(xpathatmgroup)
    if species:
        if not species == 'all':
            xpathatmgroup = '/fleurInput/atomGroups/atomGroup[@species = "{}"]'.format(species)
            xpathforce = '{}/force'.format(xpathatmgroup)
            xpathnocoParams = '{}/nocoParams'.format(xpathatmgroup)

    for key, val in six.iteritems(attributedict):
        if key == 'force':
            for attrib, value in val:
                xml_set_all_attribv(fleurinp_tree_copy, xpathforce, attrib, value)
        elif key == 'nocoParams':
            for attrib, value in val:
                xml_set_all_attribv(fleurinp_tree_copy, xpathnocoParams, attrib, value)
        else:
            xml_set_all_attribv(fleurinp_tree_copy, xpathatmgroup, attrib, value)

    return fleurinp_tree_copy

def set_inpchanges(fleurinp_tree_copy, change_dict):
    """
    Makes given changes directly in the inp.xml file. Afterwards
    updates the inp.xml file representation and the current inp_userchanges
    dictionary with the keys provided in the 'change_dict' dictionary.

    :param fleurinp_tree_copy: a lxml tree that represents inp.xml
    :param change_dict: a python dictionary with the keys to substitute.
                        It works like dict.update(), adding new keys and
                        overwriting existing keys.

    :returns new_tree: a lxml tree with applied changes

    An example of change_dict::

            change_dict = {'itmax' : 1,
                           'l_noco': True,
                           'ctail': False,
                           'l_ss': True}

    A full list of supported keys in the change_dict can be found in
    :py:func:`~aiida_fleur.tools.xml_util.get_inpxml_file_structure()`::

            'comment': '/fleurInput/comment',
            'relPos': '/fleurInput/atomGroups/atomGroup/relPos',
            'filmPos': '/fleurInput/atomGroups/atomGroup/filmPos',
            'absPos': '/fleurInput/atomGroups/atomGroup/absPos',
            'qss': '/fleurInput/calculationSetup/nocoParams/qss',
            'l_ss': '/fleurInput/calculationSetup/nocoParams',
            'row-1': '/fleurInput/cell/bulkLattice/bravaisMatrix',
            'row-2': '/fleurInput/cell/bulkLattice/bravaisMatrix',
            'row-3': '/fleurInput/cell/bulkLattice/bravaisMatrix',
            'a1': '/fleurInput/cell/filmLattice/a1',  # switches once
            'dos': '/fleurInput/output',
            'band': '/fleurInput/output',
            'secvar': '/fleurInput/calculationSetup/expertModes',
            'ctail': '/fleurInput/calculationSetup/coreElectrons',
            'frcor': '/fleurInput/calculationSetup/coreElectrons',
            'l_noco': '/fleurInput/calculationSetup/magnetism',
            'l_J': '/fleurInput/calculationSetup/magnetism',
            'swsp': '/fleurInput/calculationSetup/magnetism',
            'lflip': '/fleurInput/calculationSetup/magnetism',
            'off': '/fleurInput/calculationSetup/soc',
            'spav': '/fleurInput/calculationSetup/soc',
            'l_soc': '/fleurInput/calculationSetup/soc',
            'soc66': '/fleurInput/calculationSetup/soc',
            'pot8': '/fleurInput/calculationSetup/expertModes',
            'eig66': '/fleurInput/calculationSetup/expertModes',
            'l_f': '/fleurInput/calculationSetup/geometryOptimization',
            'gamma': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
            'gauss': '',
            'tria': '',
            'invs': '',
            'zrfs': '',
            'vchk': '/fleurInput/output/checks',
            'cdinf': '/fleurInput/output/checks',
            'disp': '/fleurInput/output/checks',
            'vacdos': '/fleurInput/output',
            'integ': '/fleurInput/output/vacuumDOS',
            'star': '/fleurInput/output/vacuumDOS',
            'iplot': '/fleurInput/output/plotting',
            'score': '/fleurInput/output/plotting',
            'plplot': '/fleurInput/output/plotting',
            'slice': '/fleurInput/output',
            'pallst': '/fleurInput/output/chargeDensitySlicing',
            'form66': '/fleurInput/output/specialOutput',
            'eonly': '/fleurInput/output/specialOutput',
            'bmt': '/fleurInput/output/specialOutput',
            'relativisticCorrections': '/fleurInput/xcFunctional',
            'calculate': '/fleurInput/atomGroups/atomGroup/force',
            'flipSpin': '/fleurInput/atomSpecies/species',
            'Kmax': '/fleurInput/calculationSetup/cutoffs',
            'Gmax': '/fleurInput/calculationSetup/cutoffs',
            'GmaxXC': '/fleurInput/calculationSetup/cutoffs',
            'numbands': '/fleurInput/calculationSetup/cutoffs',
            'itmax': '/fleurInput/calculationSetup/scfLoop',
            'minDistance': '/fleurInput/calculationSetup/scfLoop',
            'maxIterBroyd': '/fleurInput/calculationSetup/scfLoop',
            'imix': '/fleurInput/calculationSetup/scfLoop',
            'alpha': '/fleurInput/calculationSetup/scfLoop',
            'spinf': '/fleurInput/calculationSetup/scfLoop',
            'kcrel': '/fleurInput/calculationSetup/coreElectrons',
            'jspins': '/fleurInput/calculationSetup/magnetism',
            'theta': '/fleurInput/calculationSetup/soc',
            'phi': '/fleurInput/calculationSetup/soc',
            'gw': '/fleurInput/calculationSetup/expertModes',
            'lpr': '/fleurInput/calculationSetup/expertModes',
            'isec1': '/fleurInput/calculationSetup/expertModes',
            'forcemix': '/fleurInput/calculationSetup/geometryOptimization',
            'forcealpha': '/fleurInput/calculationSetup/geometryOptimization',
            'force_converged': '/fleurInput/calculationSetup/geometryOptimization',
            'qfix': '/fleurInput/calculationSetup/geometryOptimization',
            'epsdisp': '/fleurInput/calculationSetup/geometryOptimization',
            'epsforce': '/fleurInput/calculationSetup/geometryOptimization',
            'valenceElectrons': '/fleurInput/calculationSetup/bzIntegration',
            'mode': '/fleurInput/calculationSetup/bzIntegration',
            'fermiSmearingEnergy': '/fleurInput/calculationSetup/bzIntegration',
            'nx': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
            'ny': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
            'nz': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
            'count': '/fleurInput/calculationSetup/kPointCount',
            'ellow': '/fleurInput/calculationSetup/energyParameterLimits',
            'elup': '/fleurInput/calculationSetup',
            'filename': '/fleurInput/cell/symmetryFile',
            'scale': '/fleurInput/cell/bulkLattice',
            'ndir': '/fleurInput/output/densityOfStates',
            'minEnergy': '/fleurInput/output/densityOfStates',
            'maxEnergy': '/fleurInput/output/densityOfStates',
            'sigma': ' /fleurInput/output/densityOfStates',
            'layers': '/fleurInput/output/vacuumDOS',
            'nstars': '/fleurInput/output/vacuumDOS',
            'locx1': '/fleurInput/output/vacuumDOS',
            'locy1': '/fleurInput/output/vacuumDOS',
            'locx2': '/fleurInput/output/vacuumDOS',
            'locy2': '/fleurInput/output/vacuumDOS',
            'nstm': '/fleurInput/output/vacuumDOS',
            'tworkf': '/fleurInput/output/vacuumDOS',
            'numkpt': '/fleurInput/output/chargeDensitySlicing',
            'minEigenval': '/fleurInput/output/chargeDensitySlicing',
            'maxEigenval': '/fleurInput/output/chargeDensitySlicing',
            'nnne': '/fleurInput/output/chargeDensitySlicing',
            'dVac': '/fleurInput/cell/filmLattice',
            'dTilda': '/fleurInput/cell/filmLattice',
            'xcFunctional': '/fleurInput/xcFunctional/name',  # other_attributes_more
            'name': {'/fleurInput/constantDefinitions', '/fleurInput/xcFunctional',
                    '/fleurInput/atomSpecies/species'},
            'value': '/fleurInput/constantDefinitions',
            'element': '/fleurInput/atomSpecies/species',
            'atomicNumber': '/fleurInput/atomSpecies/species',
            'coreStates': '/fleurInput/atomSpecies/species',
            'magMom': '/fleurInput/atomSpecies/species',
            'radius': '/fleurInput/atomSpecies/species/mtSphere',
            'gridPoints': '/fleurInput/atomSpecies/species/mtSphere',
            'logIncrement': '/fleurInput/atomSpecies/species/mtSphere',
            'lmax': '/fleurInput/atomSpecies/species/atomicCutoffs',
            'lnonsphr': '/fleurInput/atomSpecies/species/atomicCutoffs',
            's': '/fleurInput/atomSpecies/species/energyParameters',
            'p': '/fleurInput/atomSpecies/species/energyParameters',
            'd': '/fleurInput/atomSpecies/species/energyParameters',
            'f': '/fleurInput/atomSpecies/species/energyParameters',
            'type': '/fleurInput/atomSpecies/species/lo',
            'l': '/fleurInput/atomSpecies/species/lo',
            'n': '/fleurInput/atomSpecies/species/lo',
            'eDeriv': '/fleurInput/atomSpecies/species/lo',
            'species': '/fleurInput/atomGroups/atomGroup',
            'relaxXYZ': '/fleurInput/atomGroups/atomGroup/force'

    """
    tree = fleurinp_tree_copy
    # apply changes to etree
    xmlinpstructure = get_inpxml_file_structure()
    new_tree = write_new_fleur_xmlinp_file(tree, change_dict, xmlinpstructure)

    return new_tree

def shift_value(fleurinp_tree_copy, change_dict, mode='abs'):
    """
    Shifts numertical values of some tags directly in the inp.xml file.

    :param fleurinp_tree_copy: a lxml tree that represents inp.xml
    :param change_dict: a python dictionary with the keys to shift.
    :param mode: 'abs' if change given is absolute, 'rel' if relative

    :returns new_tree: a lxml tree with shifted values

    An example of change_dict::

            change_dict = {'itmax' : 1, 'dVac': -0.123}
    """
    xmlinpstructure = get_inpxml_file_structure()
    all_attrib_xpath = xmlinpstructure[12]
    float_attributes_once = xmlinpstructure[4]
    int_attributes_once = xmlinpstructure[3]

    change_to_write = {}

    for key, value_given in six.iteritems(change_dict):
        if key not in float_attributes_once and key not in int_attributes_once:
            raise ValueError('Given attribute name either does not exist or is not floar or int')

        key_path = all_attrib_xpath[key]

        old_val = eval_xpath2(fleurinp_tree_copy, '/@'.join([key_path, key]))

        if not old_val:
            print('Can not find {} attribute in the inp.xml, skip it'.format(key))
            continue
        else:
            old_val = float(old_val[0])

        if mode == 'rel':
            value = value_given * old_val
        elif mode == 'abs':
            value = value_given + old_val
        else:
            raise ValueError("Mode should be 'res' or 'abs' only")

        if key in float_attributes_once:
            change_to_write[key] = value
        elif key in int_attributes_once:
            if not value.is_integer():
                raise ValueError('You are trying to write a float to an integer attribute')
            change_to_write[key] = int(value)

    new_tree = set_inpchanges(fleurinp_tree_copy, change_to_write)
    return new_tree

def add_num_to_att(xmltree, xpathn, attributename, set_val, mode='abs', occ=None):
    """
    Routine adds something to the value of an attribute in the xml file (should be a number here)
    This is a lower-level version of :func:`~aiida_fleur.tools.xml_util.shift_value()` which
    allows one to specife an arbitrary xml path.

    :param: an etree a xpath from root to the attribute and the attribute value
    :param xpathn: an xml path to the attribute to change
    :param attributename: a name of the attribute to change
    :param set_val: a value to be added/multiplied to the previous value
    :param mode: 'abs' if to add set_val, 'rel' if multiply
    :param occ: a list of integers specifying number of occurrence to be set

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: add_num_to_add(tree, '/fleurInput/bzIntegration', 'valenceElectrons', '1')
             add_num_to_add(tree, '/fleurInput/bzIntegration', 'valenceElectrons', '1.1', mode='rel')
    """

    if occ is None:
        occ = [0]

    # get attribute, add or multiply
    # set attribute
    attribval_node = eval_xpath(xmltree, xpathn)
    # do some checks..
    attribval = get_xml_attribute(attribval_node, attributename)
    print(attribval)
    if attribval:
        if mode == 'abs':
            newattribv = float(attribval) + float(set_val)
        elif mode == 'rel':
            newattribv = float(attribval) * float(set_val)
        else:
            pass
            # unknown mode

        xml_set_attribv_occ(xmltree, xpathn, attributename, newattribv, occ=[0], create=False)
    else:
        pass
        # something was wrong, ...
    return xmltree

def set_nkpts(fleurinp_tree_copy, count, gamma):
    """
    Sets a k-point mesh directly into inp.xml

    :param fleurinp_tree_copy: a lxml tree that represents inp.xml
    :param count: number of k-points
    :param gamma: a fortran-type boolean that controls if the gamma-point should be included
                    in the k-point mesh

    :returns new_tree: a lxml tree with applied changes
    """

    kpointlist_xpath = '/fleurInput/calculationSetup/bzIntegration/kPointList'
    #kpoint_xpath = '/fleurInput/calculationSetup/bzIntegration/kPoint*'

    tree = fleurinp_tree_copy
    new_kpo = etree.Element(
        'kPointCount',
        count="{}".format(count),
        gamma="{}".format(gamma))
    new_tree = replace_tag(tree, kpointlist_xpath, new_kpo)

    return new_tree


####### XML GETTERS #########
# TODO parser infos do not really work, might need to be returned, here


def eval_xpath(node, xpath, parser_info=None):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.
    If seferal paths are found, return a list. If only one - returns the value.

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns either nodes, or attributes, or text
    """
    if parser_info is None:
        parser_info = {'parser_warnings': []}
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        parser_info['parser_warnings'].append('There was a XpathEvalError on the xpath: {} \n'
                                              'Either it does not exist, or something is wrong'
                                              ' with the expression.'.format(xpath))
        # TODO maybe raise an error again to catch in upper routine, to know where exactly
        return []
    if len(return_value) == 1:
        return return_value[0]
    else:
        return return_value


def eval_xpath2(node, xpath, parser_info=None):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.
    Always return a list.

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns a node list
    """
    if parser_info is None:
        parser_info = {'parser_warnings': []}
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        parser_info['parser_warnings'].append('There was a XpathEvalError on the xpath: {} \n'
                                              'Either it does not exist, or something is wrong'
                                              'with the expression.'.format(xpath))
        # TODO maybe raise an error again to catch in upper routine, to know where exactly
        return []
    return return_value


def eval_xpath3(node, xpath, create=False, place_index=None, tag_order=None):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.
    If create == True, creates a tag

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns always a node list
    """
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        message = (
            'There was a XpathEvalError on the xpath: {} \n Either it does '
            'not exist, or something is wrong with the expression.'
            ''.format(xpath))
        raise etree.XPathEvalError(message)

    if return_value == []:
        if create:
            x_pieces = [e for e in xpath.split('/') if e != ""]
            #x_pieces = xpath.split('/')
            xpathn = ''
            for piece in x_pieces[:-1]:
                xpathn = xpathn + '/' + piece
            # this is REKURSIV! since create tag calls eval_xpath3
            create_tag(node, xpathn, x_pieces[-1], create=create,
                       place_index=place_index, tag_order=tag_order)
            return_value = node.xpath(xpath)
            return return_value
        else:
            return return_value
    else:
        return return_value


def get_xml_attribute(node, attributename, parser_info_out=None):
    """
    Get an attribute value from a node.

    :params node: a node from etree
    :params attributename: a string with the attribute name.
    :returns: either attributevalue, or None
    """
    if parser_info_out is None:
        parser_info_out = {'parser_warnings': []}

    if etree.iselement(node):
        attrib_value = node.get(attributename)
        if attrib_value:
            return attrib_value
        else:
            if parser_info_out:
                parser_info_out['parser_warnings'].append(
                    'Tried to get attribute: "{}" from element {}.\n '
                    'I recieved "{}", maybe the attribute does not exist'
                    ''.format(attributename, node, attrib_value))
            else:
                print((
                    'Can not get attributename: "{}" from node "{}", '
                    'because node is not an element of etree.'
                    ''.format(attributename, node)))
            return None
    else:  # something doesn't work here, some nodes get through here
        if parser_info_out:
            parser_info_out['parser_warnings'].append(
                'Can not get attributename: "{}" from node "{}", '
                'because node is not an element of etree.'
                ''.format(attributename, node))
        else:
            print(('Can not get attributename: "{}" from node "{}", '
                   'because node is not an element of etree.'
                   ''.format(attributename, node)))
        return None


# TODO this has to be done better. be able to write tags and
# certain attributes of attributes that occur possible more then once.
# HINT: This is not really used anymore. use fleurinpmodifier
def write_new_fleur_xmlinp_file(inp_file_xmltree, fleur_change_dic, xmlinpstructure):
    """
    This modifies the xml-inp file. Makes all the changes wanted by
    the user or sets some default values for certain modes

    :params inp_file_xmltree: xml-tree of the xml-inp file
    :params fleur_change_dic: dictionary {attrib_name : value} with all the wanted changes.

    :returns: an etree of the xml-inp file with changes.
    """
    # TODO rename, name is misleaded just changes the tree.
    xmltree_new = inp_file_xmltree

    pos_switch_once = xmlinpstructure[0]
    pos_switch_several = xmlinpstructure[1]
    pos_attrib_once = xmlinpstructure[2]
    pos_float_attributes_once = xmlinpstructure[4]
    pos_attrib_several = xmlinpstructure[6]
    pos_int_attributes_several = xmlinpstructure[7]
    pos_text = xmlinpstructure[11]
    pos_xpaths = xmlinpstructure[12]
    expertkey = xmlinpstructure[13]

    for key in fleur_change_dic:
        if key in pos_switch_once:
            # TODO: a test here if path is plausible and if exist
            # ggf. create tags and key.value is 'T' or 'F' if not convert,
            # if garbage, exception
            # convert user input into 'fleurbool'
            fleur_bool = convert_to_fortran_bool(fleur_change_dic[key])

            xpath_set = pos_xpaths[key]
            # TODO: check if something in setup is inconsitent?
            xml_set_first_attribv(xmltree_new, xpath_set, key, fleur_bool)

        elif key in pos_attrib_once:
            # TODO: same here, check existance and plausiblility of xpath
            xpath_set = pos_xpaths[key]
            if key in pos_float_attributes_once:
                newfloat = '{:.10f}'.format(fleur_change_dic[key])
                xml_set_first_attribv(xmltree_new, xpath_set, key, newfloat)
            elif key == 'xcFunctional':
                xml_set_first_attribv(xmltree_new, xpath_set, 'name', fleur_change_dic[key])
            else:
                xml_set_first_attribv(xmltree_new, xpath_set, key, fleur_change_dic[key])
        elif key in pos_text:
            # can be several times, therefore check
            xpath_set = pos_xpaths[key]
            xml_set_text(xmltree_new, xpath_set, fleur_change_dic[key])
        else:
            raise InputValidationError(
                "You try to set the key:'{}' to : '{}', but the key is unknown"
                " to the fleur plug-in".format(key, fleur_change_dic[key]))
    return xmltree_new


# TODO: maybe it is possible to use the xml, schema to dict libary of the QE people.
# So far it does not seem to do what we need.
def inpxml_todict(parent, xmlstr):
    """
    Recursive operation which transforms an xml etree to
    python nested dictionaries and lists.
    Decision to add a list is if the tag name is in the given list tag_several

    :param parent: some xmltree, or xml element
    :param xmlstr: structure/layout of the xml file in xmlstr is tags_several:
                   a list of the tags, which should be converted to a list, not
                   a dictionary(because they are known to occur more often, and
                   want to be accessed in a list later.

    :return: a python dictionary
    """

    xmlstructure = xmlstr
    pos_switch_once1 = xmlstructure[0]
    pos_switch_several1 = xmlstructure[1]
    int_attributes_once1 = xmlstructure[3]
    float_attributes_once1 = xmlstructure[4]
    string_attributes_once1 = xmlstructure[5]
    int_attributes_several1 = xmlstructure[7]
    float_attributes_several1 = xmlstructure[8]
    string_attributes_several1 = xmlstructure[9]
    tags_several1 = xmlstructure[10]
    pos_text1 = xmlstructure[11]

    return_dict = {}
    if list(parent.items()):
        return_dict = dict(list(parent.items()))
        # Now we have to convert lazy fortan style into pretty things for the Database
        for key in return_dict:
            if key in pos_switch_once1 or (key in pos_switch_several1):
                return_dict[key] = convert_from_fortran_bool(return_dict[key])
            elif key in int_attributes_once1 or (key in int_attributes_several1):
                # TODO int several
                try:
                    return_dict[key] = int(return_dict[key])
                except ValueError:
                    pass
            elif key in float_attributes_once1 or (key in float_attributes_several1):
                # TODO pressision?
                try:
                    return_dict[key] = float(return_dict[key])
                except ValueError:
                    pass
            elif key in string_attributes_once1 or (key in string_attributes_several1):
                # TODO What attribute shall be set? all, one or several specific onces?
                return_dict[key] = str(return_dict[key])
            elif key in pos_text1:
                # Text is done by below check (parent.text)
                pass
            else:
                pass
                # this key is not know to plug-in TODO maybe make this a method
                # of the parser and log this as warning, or add here make a log
                # list, to which you always append messages, pass them back to
                # the parser, who locks it then
                # raise TypeError("Parser wanted to convert the key:'{}' with
                # value '{}', from the inpxml file but the key is unknown to the
                # fleur plug-in".format(key, return_dict[key]))

    if parent.text:  # TODO more detal, exp: relPos
        # has text, but we don't want all the '\n' s and empty stings in the database
        if parent.text.strip() != '':  # might not be the best solution
            # set text
            return_dict = parent.text.strip()

    firstocc = True
    for element in parent:
        if element.tag in tags_several1:
            # make a list, otherwise the tag will be overwritten in the dict
            if firstocc:  # is this the first occurence?
                # create a list
                return_dict[element.tag] = []
                return_dict[element.tag].append(inpxml_todict(element, xmlstructure))
                firstocc = False
            else:  # occured before, a list already exists, therefore just add
                return_dict[element.tag].append(inpxml_todict(element, xmlstructure))
        else:
            # make dict
            return_dict[element.tag] = inpxml_todict(element, xmlstructure)

    return return_dict


# This is probably only used to represent the whole inp.xml in the database for the fleurinpData attributes
# TODO this should be replaced by something else, maybe a class. that has a method to return certain
# list of possible xpaths from a schema file, or to validate a certain xpath expression and
# to allow to get SINGLE xpaths for certain attrbiutes.
#  akk: tell me where 'DOS' is
# This might not be back compatible... i.e a certain plugin version will by this design only work
#  with certain schema version
def get_inpxml_file_structure():
    """
    This routine returns the structure/layout of the 'inp.xml' file.

    Basicly the plug-in should know from this routine, what things are allowed
    to be set and where, i.e all attributes and their xpaths.
    As a developer make sure to use this routine always of you need information
    about the inp.xml file structure.
    Therefore, this plug-in should be easy to adjust to other codes with xml
    files as input files. Just rewrite this routine.

    For now the structure of the xmlinp file for fleur is hardcoded.
    If big changes are in the 'inp.xml' file, maintain this routine.
    TODO: Maybe this is better done, by reading the xml schema datei instead.
    And maybe it should also work without the schema file, do we want this?

    :param Nothing: TODO xml schema

    :return all_switches_once: list of all switches ('T' or 'F') which are allowed to be set
    :return all_switches_several: list of all switches ('T' or 'F') which are allowed to be set
    :return other_attributes_once: list of all attributes, which occur just once (can be tested)
    :return other_attributes_several: list of all attributes, which can occur more then once
    :return all_text: list of all text of tags, which can be set
    :return all_attrib_xpath:
                              dictonary (attrib, xpath), of all possible attributes
                              with their xpath expression for the xmp inp

    :return expertkey:
                       keyname (should not be in any other list), which can be
                       used to set anything in the file, by hand,
                       (for experts, and that plug-in does not need to be directly maintained if
                       xmlinp gets a new switch)
    """

    # All attributes (allowed to change?)

    # switches can be 'T' ot 'F' # TODO: alphabetical sorting
    all_switches_once = (
        'dos', 'band', 'secvar', 'ctail', 'frcor', 'l_noco',
        'ctail', 'swsp', 'lflip', 'off', 'spav', 'l_soc', 'soc66', 'pot8',
        'eig66', 'gamma', 'gauss', 'tria', 'invs', 'invs2', 'zrfs', 'vchk', 'cdinf',
        'disp', 'vacdos', 'integ', 'star', 'score', 'plplot', 'slice',
        'pallst', 'form66', 'eonly', 'bmt', 'relativisticCorrections', 'l_J', 'l_f', 'l_ss')

    all_switches_several = ('calculate', 'flipSpin')

    int_attributes_once = ('numbands', 'itmax', 'maxIterBroyd', 'kcrel', 'jspins',
                           'gw', 'isec1', 'nx', 'ny', 'nz', 'ndir', 'layers',
                           'nstars', 'nstm', 'iplot', 'numkpt', 'nnne', 'lpr', 'count', 'qfix')

    float_attributes_once = ('Kmax', 'Gmax', 'GmaxXC', 'alpha', 'spinf', 'minDistance', 'theta',
                             'phi', 'epsdisp', 'epsforce',
                             'valenceElectrons', 'fermiSmearingEnergy', 'ellow',
                             'elup', 'scale', 'dTilda', 'dVac', 'minEnergy',
                             'maxEnergy', 'sigma', 'locx1', 'locy1', 'locx2',
                             'locy2', 'tworkf', 'minEigenval', 'maxEigenval',
                             'forcealpha', 'force_converged')

    string_attributes_once = ('imix', 'mode', 'filename', 'latnam', 'spgrp',
                              'xcFunctional', 'fleurInputVersion', 'species', 'forcemix')

    other_attributes_once = tuple(
        list(int_attributes_once) +
        list(float_attributes_once) +
        list(string_attributes_once))
    other_attributes_once1 = (
        'isec1', 'Kmax', 'Gmax', 'GmaxXC', 'numbands', 'itmax', 'maxIterBroyd',
        'imix', 'alpha', 'spinf', 'minDistance',
        'kcrel', 'jspins', 'theta', 'phi', 'gw', 'lpr',
        'epsdisp', 'epsforce', 'valenceElectrons', 'mode',
        'gauss', 'fermiSmearingEnergy', 'nx', 'ny', 'nz', 'ellow', 'elup',
        'filename', 'scale', 'dTilda', 'dVac', 'ndir', 'minEnergy', 'maxEnergy',
        'sigma', 'layers', 'nstars', 'locx1', 'locy1', 'locx2', 'locy2', 'nstm',
        'tworkf', 'numkpt', 'minEigenval', 'maxEigenval', 'nnne')

    int_attributes_several = ('atomicNumber', 'gridPoints', 'lmax', 'lnonsphr',
                              's', 'p', 'd', 'f', 'l', 'n', 'eDeriv', 'coreStates')
    float_attributes_several = ('value', 'magMom', 'radius', 'logIncrement')
    string_attributes_several = ('name', 'element', 'coreStates', 'type', 'relaxXYZ')
    other_attributes_several = (
        'name', 'value', 'element', 'atomicNumber', 'coreStates', 'magMom',
        'radius', 'gridPoints', 'logIncrement', 'lmax', 'lnonsphr', 's', 'p',
        'd', 'f', 'species', 'type', 'coreStates', 'l', 'n', 'eDeriv', 'relaxXYZ')

    # when parsing the xml file to a dict, these tags should become
    # list(sets, or tuples) instead of dictionaries.
    tags_several = ('atomGroup', 'relPos', 'absPos', 'filmPos',
                    'species', 'kPoint', 'lo', 'stateOccupation')

    all_text = {'comment': 1, 'relPos': 3, 'filmPos': 3, 'absPos': 3,
                'row-1': 3, 'row-2': 3, 'row-3': 3, 'a1': 1, 'qss': 3}
    # TODO all these (without comment) are floats, or float tuples.
    # Should be converted to this in the databas
    # changing the Bravais matrix should rather not be allowed I guess

    # all attribute xpaths

    # text xpaths(coordinates, bravaisMatrix)
    # all switches once, several, all attributes once, several
    all_attrib_xpath = {  # text
        'comment': '/fleurInput/comment',
        'relPos': '/fleurInput/atomGroups/atomGroup/relPos',
        'filmPos': '/fleurInput/atomGroups/atomGroup/filmPos',
        'absPos': '/fleurInput/atomGroups/atomGroup/absPos',
        'qss': '/fleurInput/calculationSetup/nocoParams/qss',
        'l_ss': '/fleurInput/calculationSetup/nocoParams',
        'row-1': '/fleurInput/cell/bulkLattice/bravaisMatrix',
        'row-2': '/fleurInput/cell/bulkLattice/bravaisMatrix',
        'row-3': '/fleurInput/cell/bulkLattice/bravaisMatrix',
        'a1': '/fleurInput/cell/filmLattice/a1',  # switches once
        'dos': '/fleurInput/output',
        'band': '/fleurInput/output',
        'secvar': '/fleurInput/calculationSetup/expertModes',
        'ctail': '/fleurInput/calculationSetup/coreElectrons',
        'frcor': '/fleurInput/calculationSetup/coreElectrons',
        'l_noco': '/fleurInput/calculationSetup/magnetism',
        'l_J': '/fleurInput/calculationSetup/magnetism',
        'swsp': '/fleurInput/calculationSetup/magnetism',
        'lflip': '/fleurInput/calculationSetup/magnetism',
        'off': '/fleurInput/calculationSetup/soc',
        'spav': '/fleurInput/calculationSetup/soc',
        'l_soc': '/fleurInput/calculationSetup/soc',
        'soc66': '/fleurInput/calculationSetup/soc',
        'pot8': '/fleurInput/calculationSetup/expertModes',
        'eig66': '/fleurInput/calculationSetup/expertModes',
        'l_f': '/fleurInput/calculationSetup/geometryOptimization',
        'gamma': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        # 'invs': '',
        # 'zrfs': '',
        'vchk': '/fleurInput/output/checks',
        'cdinf': '/fleurInput/output/checks',
        'disp': '/fleurInput/output/checks',
        'vacdos': '/fleurInput/output',
        'integ': '/fleurInput/output/vacuumDOS',
        'star': '/fleurInput/output/vacuumDOS',
        'iplot': '/fleurInput/output/plotting',
        'score': '/fleurInput/output/plotting',
        'plplot': '/fleurInput/output/plotting',
        'slice': '/fleurInput/output',
        'pallst': '/fleurInput/output/chargeDensitySlicing',
        'form66': '/fleurInput/output/specialOutput',
        'eonly': '/fleurInput/output/specialOutput',
        'bmt': '/fleurInput/output/specialOutput',
        'relativisticCorrections': '/fleurInput/xcFunctional',  # ALL_Switches_several
        'calculate': '/fleurInput/atomGroups/atomGroup/force',
        'flipSpin': '/fleurInput/atomSpecies/species',  # other_attributes_once
        'Kmax': '/fleurInput/calculationSetup/cutoffs',
        'Gmax': '/fleurInput/calculationSetup/cutoffs',
        'GmaxXC': '/fleurInput/calculationSetup/cutoffs',
        'numbands': '/fleurInput/calculationSetup/cutoffs',
        'itmax': '/fleurInput/calculationSetup/scfLoop',
        'minDistance': '/fleurInput/calculationSetup/scfLoop',
        'maxIterBroyd': '/fleurInput/calculationSetup/scfLoop',
        'imix': '/fleurInput/calculationSetup/scfLoop',
        'alpha': '/fleurInput/calculationSetup/scfLoop',
        'spinf': '/fleurInput/calculationSetup/scfLoop',
        'kcrel': '/fleurInput/calculationSetup/coreElectrons',
        'jspins': '/fleurInput/calculationSetup/magnetism',
        'theta': '/fleurInput/calculationSetup/soc',
        'phi': '/fleurInput/calculationSetup/soc',
        'gw': '/fleurInput/calculationSetup/expertModes',
        'lpr': '/fleurInput/calculationSetup/expertModes',
        'isec1': '/fleurInput/calculationSetup/expertModes',
        'forcemix': '/fleurInput/calculationSetup/geometryOptimization',
        'forcealpha': '/fleurInput/calculationSetup/geometryOptimization',
        'force_converged': '/fleurInput/calculationSetup/geometryOptimization',
        'qfix': '/fleurInput/calculationSetup/geometryOptimization',
        'epsdisp': '/fleurInput/calculationSetup/geometryOptimization',
        'epsforce': '/fleurInput/calculationSetup/geometryOptimization',
        'valenceElectrons': '/fleurInput/calculationSetup/bzIntegration',
        'mode': '/fleurInput/calculationSetup/bzIntegration',
        'fermiSmearingEnergy': '/fleurInput/calculationSetup/bzIntegration',
        'nx': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'ny': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'nz': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'count': '/ fleurInput/calculationSetup/bzIntegration/kPointList',
        'ellow': '/fleurInput/calculationSetup/energyParameterLimits',
        'elup': '/fleurInput/calculationSetup/energyParameterLimits',
        #'filename': '/fleurInput/cell/symmetryFile',
        'scale': '/fleurInput/cell/bulkLattice',
        # 'film_scale': '/fleurInput/cell/filmLattice',
        'ndir': '/fleurInput/output/densityOfStates',
        'minEnergy': '/fleurInput/output/densityOfStates',
        'maxEnergy': '/fleurInput/output/densityOfStates',
        'sigma': ' /fleurInput/output/densityOfStates',
        'layers': '/fleurInput/output/vacuumDOS',
        'nstars': '/fleurInput/output/vacuumDOS',
        'locx1': '/fleurInput/output/vacuumDOS',
        'locy1': '/fleurInput/output/vacuumDOS',
        'locx2': '/fleurInput/output/vacuumDOS',
        'locy2': '/fleurInput/output/vacuumDOS',
        'nstm': '/fleurInput/output/vacuumDOS',
        'tworkf': '/fleurInput/output/vacuumDOS',
        'numkpt': '/fleurInput/output/chargeDensitySlicing',
        'minEigenval': '/fleurInput/output/chargeDensitySlicing',
        'maxEigenval': '/fleurInput/output/chargeDensitySlicing',
        'nnne': '/fleurInput/output/chargeDensitySlicing',
        'dVac': '/fleurInput/cell/filmLattice',
        'dTilda': '/fleurInput/cell/filmLattice',
        'xcFunctional': '/fleurInput/xcFunctional',  # other_attributes_more
        # 'name': {'/fleurInput/constantDefinitions', '/fleurInput/xcFunctional',
        #          '/fleurInput/atomSpecies/species'},
        # 'value': '/fleurInput/constantDefinitions',
        'element': '/fleurInput/atomSpecies/species',
        'atomicNumber': '/fleurInput/atomSpecies/species',
        'coreStates': '/fleurInput/atomSpecies/species',
        'magMom': '/fleurInput/atomSpecies/species',
        'radius': '/fleurInput/atomSpecies/species/mtSphere',
        'gridPoints': '/fleurInput/atomSpecies/species/mtSphere',
        'logIncrement': '/fleurInput/atomSpecies/species/mtSphere',
        'lmax': '/fleurInput/atomSpecies/species/atomicCutoffs',
        'lnonsphr': '/fleurInput/atomSpecies/species/atomicCutoffs',
        's': '/fleurInput/atomSpecies/species/energyParameters',
        'p': '/fleurInput/atomSpecies/species/energyParameters',
        'd': '/fleurInput/atomSpecies/species/energyParameters',
        'f': '/fleurInput/atomSpecies/species/energyParameters',
        'type': '/fleurInput/atomSpecies/species/lo',
        'l': '/fleurInput/atomSpecies/species/lo',
        'n': '/fleurInput/atomSpecies/species/lo',
        'eDeriv': '/fleurInput/atomSpecies/species/lo',
        'species': '/fleurInput/atomGroups/atomGroup',
        'relaxXYZ': '/fleurInput/atomGroups/atomGroup/force'
    }

    all_tag_xpaths = (
        '/fleurInput/constantDefinitions',
        '/fleurInput/calculationSetup',
        '/fleurInput/calculationSetup/cutoffs',
        '/fleurInput/calculationSetup/scfLoop',
        '/fleurInput/calculationSetup/coreElectrons',
        '/fleurInput/calculationSetup/magnetism',
        '/fleurInput/calculationSetup/soc',
        '/fleurInput/calculationSetup/expertModes',
        '/fleurInput/calculationSetup/geometryOptimization',
        '/fleurInput/calculationSetup/bzIntegration',
        '/fleurInput/calculationSetup/kPointMesh',
        '/fleurInput/cell/symmetry',
        '/fleurInput/cell/bravaisMatrix',
        '/fleurInput/calculationSetup/nocoParams',
        '/fleurInput/xcFunctional',
        '/fleurInput/xcFunctional/xcParams',
        '/fleurInput/atomSpecies/species',
        '/fleurInput/atomSpecies/species/mtSphere',
        '/fleurInput/atomSpecies/species/atomicCutoffs',
        '/fleurInput/atomSpecies/species/energyParameters',
        '/fleurInput/atomSpecies/species/coreConfig',
        '/fleurInput/atomSpecies/species/coreOccupation',
        '/fleurInput/atomGroups/atomGroup',
        '/fleurInput/atomGroups/atomGroup/relPos',
        '/fleurInput/atomGroups/atomGroup/absPos',
        '/fleurInput/atomGroups/atomGroup/filmPos',
        '/fleurInput/output/checks',
        '/fleurInput/output/densityOfStates',
        '/fleurInput/output/vacuumDOS',
        '/fleurInput/output/plotting',
        '/fleurInput/output/chargeDensitySlicing',
        '/fleurInput/output/specialOutput'
    )

    expertkey = 'other'
    returnlist = (all_switches_once,
                  all_switches_several,
                  other_attributes_once,
                  int_attributes_once,
                  float_attributes_once,
                  string_attributes_once,
                  other_attributes_several,
                  int_attributes_several,
                  float_attributes_several,
                  string_attributes_several,
                  tags_several,
                  all_text,
                  all_attrib_xpath,
                  expertkey)
    return returnlist

def clear_xml(tree):
    """
    Removes comments and executes xinclude tags of an
    xml tree.

    :param tree: an xml-tree which will be processes
    :return cleared_tree: an xml-tree without comments and with replaced xinclude tags
    """
    import copy

    cleared_tree = copy.deepcopy(tree)

    # replace XInclude parts to validate against schema
    cleared_tree.xinclude()

    # remove comments from inp.xml
    comments = cleared_tree.xpath('//comment()')
    for comment in comments:
        com_parent = comment.getparent()
        com_parent.remove(comment)

    return cleared_tree



#!/usr/bin/env python

from __future__ import absolute_import
from setuptools import setup, find_packages
import json

if __name__ == '__main__':
    # Provide static information in setup.json
    # such that it can be discovered automatically
    with open('setup.json', 'r') as info:
        kwargs = json.load(info)
    setup(
        packages=find_packages(exclude='aiida'),
        **kwargs
    )



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
This contains code snippets and utility useful for dealing with parameter data nodes
commonly used by the fleur plugin and workflows
"""
from __future__ import print_function
from __future__ import absolute_import
import six

def extract_elementpara(parameter_dict, element):
    """
    :param parameter_dict: python dict, parameter node for inpgen
    :param element: string, i.e 'W'

    :return: python dictionary, parameter node which contains only
                                the atom parameters for the given element
    """
    element_para_dict = {}
    for key, val in six.iteritems(parameter_dict):
        if 'atom' in key:
            if val.get('element', '') == element:
                element_para_dict[key] = val
        else:
            element_para_dict[key] = val
    return element_para_dict


def dict_merger(dict1, dict2):
    """
    Merge recursively two nested python dictionaries.

    If key is in both digionaries tries to add the entries in both dicts.
    (merges two subdicts, adds strings and numbers together)

    :return: dict
    """
    new_dict = dict1.copy()

    if not dict1:
        return dict2
    if not dict2:
        return dict1

    keys1 = list(dict1.keys())
    keys2 = list(dict2.keys())

    # add uncommon
    for key in keys2:
        if key not in keys1:
            new_dict[key] = dict2[key]

    # merge common
    for key, val in six.iteritems(dict1):
        if isinstance(val, dict):
            new_dict[key] = dict_merger(val, dict2.get(key, {}))
        elif isinstance(val, list):
            new_dict[key] = val + dict2.get(key, [])
        elif isinstance(val, str):
            new_dict[key] = val + dict2.get(key, '')
        elif isinstance(val, int):
            new_dict[key] = val + dict2.get(key, 0)
        elif isinstance(val, float):
            new_dict[key] = val + dict2.get(key, 0.0)
        else:
            print(("don't know what to do with element : {}".format(key)))
    return new_dict



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module you find the workflow 'FleurEosWorkChain' for the calculation of
of an equation of state
"""
# TODO: print more user info
# allow different inputs, make things optional(don't know yet how)
# half number of iteration if you are close to be converged. (therefore
# one can start with 18 iterations, and if thats not enough run again 9 or something)
from __future__ import absolute_import
from __future__ import print_function
import numpy as np

import six
from six.moves import range

from aiida.orm import load_node
from aiida.orm import Float, StructureData, Dict
from aiida.engine import WorkChain, ToContext
from aiida.engine import calcfunction as cf
from aiida.common import AttributeDict

from aiida_fleur.tools.StructureData_util import rescale, rescale_nowf, is_structure
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.tools.common_fleur_wf_util import check_eos_energies


class FleurEosWorkChain(WorkChain):
    """
    This workflow calculates the equation of states of a structure.
    Calculates several unit cells with different volumes.
    A Birch_Murnaghan  equation of states fit determines the Bulk modulus and the
    groundstate volume of the cell.

    :params wf_parameters: Dict node, optional 'wf_parameters', protocol specifying parameter dict
    :params structure: StructureData node, 'structure' crystal structure
    :params calc_parameters: Dict node, optional 'calc_parameters' parameters for inpgen
    :params inpgen: Code node,
    :params fleur: Code node,


    :return output_eos_wc_para: Dict node, contains relevant output information.
                                about general succeed, fit results and so on.
    """

    _workflowversion = "0.3.5"

    _wf_default = {
        'points': 9,
        'step': 0.002,
        'guess': 1.00
    }

    @classmethod
    def define(cls, spec):
        super(FleurEosWorkChain, cls).define(spec)
        spec.expose_inputs(FleurScfWorkChain, namespace='scf', exclude=('structure', ))
        spec.input("wf_parameters", valid_type=Dict, required=False)
        spec.input("structure", valid_type=StructureData, required=True)

        spec.outline(
            cls.start,
            cls.structures,
            cls.converge_scf,
            cls.return_results
        )

        spec.output('output_eos_wc_para', valid_type=Dict)
        spec.output('output_eos_wc_structure', valid_type=StructureData)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM',
                       message="Invalid workchain parameters.")

    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        self.report("Started eos workflow version {}".format(self._workflowversion))

        self.ctx.last_calc2 = None
        self.ctx.calcs = []
        self.ctx.calcs_future = []
        self.ctx.structures = []
        self.ctx.temp_calc = None
        self.ctx.structurs_uuids = []
        self.ctx.scalelist = []
        self.ctx.volume = []
        self.ctx.volume_peratom = {}
        self.ctx.org_volume = -1  # avoid div 0
        self.ctx.labels = []
        self.ctx.successful = True
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        # TODO get all successful from convergence, if all True this

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = self._wf_default
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = 'ERROR: input wf_parameters for EOS contains extra keys: {}'.format(extra_keys)
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in six.iteritems(wf_default):
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        self.ctx.points = wf_dict.get('points', 9)
        self.ctx.step = wf_dict.get('step', 0.002)
        self.ctx.guess = wf_dict.get('guess', 1.00)
        self.ctx.serial = wf_dict.get('serial', False)  # True
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)

    def structures(self):
        """
        Creates structure data nodes with different Volume (lattice constants)
        """
        points = self.ctx.points
        step = self.ctx.step
        guess = self.ctx.guess
        startscale = guess - (points - 1) / 2 * step

        for point in range(points):
            self.ctx.scalelist.append(startscale + point * step)

        self.report('scaling factors which will be calculated:{}'.format(self.ctx.scalelist))
        self.ctx.org_volume = self.inputs.structure.get_cell_volume()
        self.ctx.structurs = eos_structures(self.inputs.structure, self.ctx.scalelist)

    def converge_scf(self):
        """
        Launch fleur_scfs from the generated structures
        """
        calcs = {}

        for i, struc in enumerate(self.ctx.structurs):
            inputs = self.get_inputs_scf()
            inputs.structure = struc
            natoms = len(struc.sites)
            label = str(self.ctx.scalelist[i])
            label_c = '|eos| fleur_scf_wc'
            description = '|FleurEosWorkChain|fleur_scf_wc|scale {}, {}'.format(label, i)
            #inputs.label = label_c
            #inputs.description = description

            self.ctx.volume.append(struc.get_cell_volume())
            self.ctx.volume_peratom[label] = struc.get_cell_volume() / natoms
            self.ctx.structurs_uuids.append(struc.uuid)

            result = self.submit(FleurScfWorkChain, **inputs)
            self.ctx.labels.append(label)
            calcs[label] = result

        return ToContext(**calcs)

    def get_inputs_scf(self):
        """
        get and 'produce' the inputs for a scf-cycle
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        return input_scf

    def return_results(self):
        """
        return the results of the calculations  (scf workchains) and do a
        Birch-Murnaghan fit for the equation of states
        """
        distancelist = []
        t_energylist = []
        t_energylist_peratom = []
        vol_peratom_success = []
        outnodedict = {}
        natoms = len(self.inputs.structure.sites)
        htr_to_ev = 27.21138602
        e_u = 'eV'

        for label in self.ctx.labels:
            calc = self.ctx[label]

            if not calc.is_finished_ok:
                message = ('One SCF workflow was not successful: {}'.format(label))
                self.ctx.warnings.append(message)
                self.ctx.successful = False
                continue

            try:
                _ = calc.outputs.output_scf_wc_para
            except KeyError:
                message = (
                    'One SCF workflow failed, no scf output node: {}.'
                    ' I skip this one.'.format(label))
                self.ctx.errors.append(message)
                self.ctx.successful = False
                continue

            outpara = calc.outputs.output_scf_wc_para.get_dict()

            t_e = outpara.get('total_energy', float('nan'))
            e_u = outpara.get('total_energy_units', 'eV')
            if e_u == 'Htr' or 'htr':
                t_e = t_e * htr_to_ev
            dis = outpara.get('distance_charge', float('nan'))
            dis_u = outpara.get('distance_charge_units')
            t_energylist.append(t_e)
            t_energylist_peratom.append(t_e / natoms)
            vol_peratom_success.append(self.ctx.volume_peratom[label])
            distancelist.append(dis)

        not_ok, an_index = check_eos_energies(t_energylist_peratom)

        if not_ok:
            message = ('Abnormality in Total energy list detected. Check '
                       'entr(ies) {}.'.format(an_index))
            hint = ('Consider refining your basis set.')
            self.ctx.info.append(hint)
            self.ctx.warnings.append(message)

        en_array = np.array(t_energylist_peratom)
        vol_array = np.array(vol_peratom_success)

        write_defaults_fit = False
        # TODO: different fits
        if len(en_array):  # for some reason just en_array does not work
            volume, bulk_modulus, bulk_deriv, residuals = birch_murnaghan_fit(en_array, vol_array)

            for i in volume, bulk_modulus, bulk_deriv, residuals:
                if issubclass(type(i), np.complex):
                    write_defaults_fit = True
            volumes = self.ctx.volume
            gs_scale = volume * natoms / self.ctx.org_volume
            if (volume * natoms < volumes[0]) or (volume * natoms > volumes[-1]):
                warn = ('Groundstate volume was not in the scaling range.')
                hint = ('Consider rerunning around point {}'.format(gs_scale))
                self.ctx.info.append(hint)
                self.ctx.warnings.append(warn)
                # TODO maybe make it a feature to rerun with centered around the gs.
        else:
            write_defaults_fit = True

        if write_defaults_fit:
            volumes = None
            gs_scale = None
            residuals = None
            volume = 0
            bulk_modulus = None
            bulk_deriv = None

        out = {'workflow_name': self.__class__.__name__,
               'workflow_version': self._workflowversion,
               'scaling': self.ctx.scalelist,
               'scaling_gs': gs_scale,
               'initial_structure': self.inputs.structure.uuid,
               'volume_gs': volume * natoms,
               'volumes': volumes,
               'volume_units': 'A^3',
               'natoms': natoms,
               'total_energy': t_energylist,
               'total_energy_units': e_u,
               'structures': self.ctx.structurs_uuids,
               'calculations': [],  # self.ctx.calcs1,
               'scf_wfs': [],  # self.converge_scf_uuids,
               'distance_charge': distancelist,
               'distance_charge_units': dis_u,
               'nsteps': self.ctx.points,
               'guess': self.ctx.guess,
               'stepsize': self.ctx.step,
               # 'fitresults' : [a, latticeconstant, c],
               # 'fit' : fit_new,
               'residuals': residuals,
               'bulk_deriv': bulk_deriv,
               'bulk_modulus': bulk_modulus * 160.217733,  # * echarge * 1.0e21,#GPa
               'bulk_modulus_units': 'GPa',
               'info': self.ctx.info,
               'warnings': self.ctx.warnings,
               'errors': self.ctx.errors
               }

        if self.ctx.successful:
            self.report('Done, Equation of states calculation complete')
        else:
            self.report(
                'Done, but something went wrong.... Probably some individual calculation failed or'
                ' a scf-cycle did not reach the desired distance.')

        outnode = Dict(dict=out)
        outnodedict['results_node'] = outnode

        # create links between all these nodes...
        outputnode_dict = create_eos_result_node(**outnodedict)
        outputnode = outputnode_dict.get('output_eos_wc_para')
        outputnode.label = 'output_eos_wc_para'
        outputnode.description = ('Contains equation of states results and information of an '
                                  'FleurEosWorkChain run.')

        returndict = {}
        returndict['output_eos_wc_para'] = outputnode

        outputstructure = outputnode_dict.get('gs_structure', None)
        if outputstructure:
            outputstructure.label = 'output_eos_wc_structure'
            outputstructure.description = ('Structure with the scaling/volume of the lowest total '
                                           'energy extracted from FleurEosWorkChain')
            outputstructure = save_structure(outputstructure)
            returndict['output_eos_wc_structure'] = outputstructure

        # create link to workchain node
        for link_name, node in six.iteritems(returndict):
            self.out(link_name, node)

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()

        return


@cf
def create_eos_result_node(**kwargs):
    """
    This is a pseudo wf, to create the right graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_eos_wc_para'] = outpara.clone()
    # copy, because we rather produce the same node twice
    # then have a circle in the database for now...
    outputdict = outpara.get_dict()
    structure = load_node(outputdict.get('initial_structure'))
    gs_scaling = outputdict.get('scaling_gs', 0)
    if gs_scaling:
        gs_structure = rescale_nowf(structure, Float(gs_scaling))
        outdict['gs_structure'] = gs_structure

    return outdict


@cf
def save_structure(structure):
    """
    Saves a structure data node
    """
    structure_return = structure.clone()
    return structure_return


def eos_structures(inp_structure, scalelist):
    """
    Creates many rescalled StructureData nodes out of a crystal structure.
    Keeps the provenance in the database.

    :param StructureData, a StructureData node (pk, sor uuid)
    :param scalelist, list of floats, scaling factors for the cell

    :returns: list of New StructureData nodes with rescalled structure, which are linked to input
              Structure
    """
    structure = is_structure(inp_structure)
    if not structure:
        # TODO: log something (test if it gets here at all)
        return None
    re_structures = []

    for scale in scalelist:
        structure_rescaled = rescale(structure, Float(scale))  # this is a wf
        re_structures.append(structure_rescaled)

    return re_structures

# pylint: disable=invalid-name


def birch_murnaghan_fit(energies, volumes):
    """
    least squares fit of a Birch-Murnaghan equation of state curve. From delta project
    containing in its columns the volumes in A^3/atom and energies in eV/atom
    # The following code is based on the source code of eos.py from the Atomic
    # Simulation Environment (ASE) <https://wiki.fysik.dtu.dk/ase/>.
    :params energies: list (numpy arrays!) of total energies eV/atom
    :params volumes: list (numpy arrays!) of volumes in A^3/atom

    #volume, bulk_modulus, bulk_deriv, residuals = Birch_Murnaghan_fit(data)
    """
    fitdata = np.polyfit(volumes[:]**(-2. / 3.), energies[:], 3, full=True)
    ssr = fitdata[1]
    sst = np.sum((energies[:] - np.average(energies[:]))**2.)

    residuals0 = ssr / sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x**(-3. / 2.)
            break

    if volume0 == 0:
        print('Error: No minimum could be found')
        exit()

    derivV2 = 4. / 9. * x**5. * deriv2(x)
    derivV3 = (-20. / 9. * x**(13. / 2.) * deriv2(x) -
               8. / 27. * x**(15. / 2.) * deriv3(x))
    bulk_modulus0 = derivV2 / x**(3. / 2.)
    bulk_deriv0 = -1 - x**(-3. / 2.) * derivV3 / derivV2

    return volume0, bulk_modulus0, bulk_deriv0, residuals0


def birch_murnaghan(volumes, volume0, bulk_modulus0, bulk_deriv0):
    """
    This evaluates the Birch Murnaghan equation of states
    """
    PV = []
    EV = []
    v0 = volume0
    bm = bulk_modulus0
    dbm = bulk_deriv0

    for vol in volumes:
        pv_val = 3 * bm / 2. * ((v0 / vol)**(7 / 3.) - (v0 / vol)**(5 / 3.)) * \
            (1 + 3 / 4. * (dbm - 4) * ((v0 / vol)**(2 / 3.) - 1))
        PV.append(pv_val)
        ev_val = 9 * bm * v0 / 16. * ((dbm * (v0 / vol)**(2 / 3.) - 1)**(3) *
                                      ((v0 / vol)**(2 / 3.) - 1)**2 * (6 - 4 * (v0 / vol)**(2 / 3.)))
        EV.append(ev_val)
    return EV, PV



from __future__ import absolute_import
import pytest

#class test_extract_corelevels():

# TODO: test in general more results outcome.
def test_extract_corelevels_outfile_allfiles():
    """
    Extracts corelevels and atomtype imformation from example out.xml file
    """
    from aiida_fleur.tools.extract_corelevels import extract_corelevels

    outxmlfiles = get_example_outxml_files()
    
    for outfile in outxmlfiles:
        corelevel, atomtypes = extract_corelevels(outfile)
        assert bool(corelevel)
        assert bool(atomtypes)

def test_extract_corelevels_outfile_interface():
    """
    Extracts corelevels and atomtype imformation from one example out.xml file
    and check the format and values of the results.
    """
    from aiida_fleur.tools.extract_corelevels import extract_corelevels

    pass
    #outfile = 
    #corelevel, atomtypes = extract_corelevels(outfile)
    #    assert bool(corelevel)
    #    assert bool(atomtypes)

'''
# this is used in extrac_corelevels, but might make sense to test extra
def test_parse_state_card_interface():
    """
    Parses a state card as written in given example out.xml files.
    The test should fail if the format of the state card in the out.xml file changes
    and/or if the interface of the method changes
    """
    from aiida_fleur.tools.extract_corelevels import parse_state_card
'''

def test_clsshifts_to_be_interface():
    """
    Tests the interface of the clsshifts_to_be method.
    Corelevel shifts with no reference given, or an empty reference given should be ignored.
    or a warning issued.
    """
    from aiida_fleur.tools.extract_corelevels import clshifts_to_be

    reference = {'W' : {'4f7/2' : [124],'4f5/2' : [], '3p3/2' : [], '3p5/2' : [10] }, 'Be' : {'1s': [117]}}
    corelevels = {'W' : {'4f7/2' : [-0.4, 0.3],'4f5/2' : [0, 0.1]},
                 'Be' : {'1s': [0,  0.3]}, 'C' : [0.0]}
    
    res = clshifts_to_be(corelevels, reference)

    expected_results = {'Be': {'1s': [117, 117.3]}, 'W': {'4f7/2': [123.6, 124.3]}}
    assert res == expected_results
    
    # expect warning clsshifts_to_be(corelevels, reference, warn=True)
    #Warning: Reference for element: 'C' not given. I ignore these.
    #Warning: Reference corelevel '4f5/2' for element: 'W' not given. I ignore these.


def get_example_outxml_files():
    """
    helper. returns all the realativ paths to the example out.xml files.
    """
    from os import listdir
    from os.path import join   
    # from top test folder
    folder_path = './files/outxml/'
    return [join(folder_path, outfile) for outfile in listdir(folder_path) if outfile.endswith(".xml")]
    




# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
    In this module you find the workflow 'FleurRelaxWorkChain' for geometry optimization.
"""
from __future__ import absolute_import
from __future__ import print_function
import copy
import numpy as np

import six

from aiida.engine import WorkChain, ToContext, while_
from aiida.engine import calcfunction as cf
from aiida.orm import load_node
from aiida.orm import StructureData, Dict
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.calculation.fleur import FleurCalculation as FleurCalc


class FleurRelaxWorkChain(WorkChain):
    """
    This workflow performs structure optimization.
    """

    _workflowversion = "0.1.3"

    _wf_default = {
        'relax_iter': 5,
        'film_distance_relaxation': False,
        'force_criterion': 0.001
    }

    @classmethod
    def define(cls, spec):
        super(FleurRelaxWorkChain, cls).define(spec)
        spec.expose_inputs(FleurScfWorkChain, namespace='scf')
        spec.input("wf_parameters", valid_type=Dict, required=False)

        spec.outline(
            cls.start,
            cls.converge_scf,
            cls.check_failure,
            while_(cls.condition)(
                cls.generate_new_fleurinp,
                cls.converge_scf,
                cls.check_failure,
            ),
            cls.get_results,
            cls.return_results,
        )

        spec.output('out', valid_type=Dict)
        spec.output('optimized_structure', valid_type=StructureData)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM',
                       message="Invalid workchain parameters.")
        spec.exit_code(350, 'ERROR_DID_NOT_RELAX',
                       message="Optimization cycle did not lead to convergence of forces.")
        spec.exit_code(351, 'ERROR_SCF_FAILED',
                       message="SCF Workchains failed for some reason.")
        spec.exit_code(352, 'ERROR_NO_RELAX_OUTPUT',
                       message="Found no relaxed structure info in the output of SCF")
        spec.exit_code(352, 'ERROR_NO_SCF_OUTPUT',
                       message="Found no SCF output")
        spec.exit_code(311, 'ERROR_VACUUM_SPILL_RELAX',
                       message='FLEUR calculation failed because an atom spilled to the'
                               'vacuum during relaxation')
        spec.exit_code(313, 'ERROR_MT_RADII_RELAX',
                       message='Overlapping MT-spheres during relaxation.')

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain
        """
        self.report('INFO: started structure relaxation workflow version {}\n'
                    ''.format(self._workflowversion))

        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []

        # Pre-initialization of some variables
        self.ctx.loop_count = 0
        self.ctx.forces = []
        self.ctx.final_cell = None
        self.ctx.final_atom_positions = None
        self.ctx.pbc = None
        self.ctx.reached_relax = True
        self.ctx.scf_res = None

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._wf_default)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = 'ERROR: input wf_parameters for Relax contains extra keys: {}'.format(
                extra_keys)
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in six.iteritems(wf_default):
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

    def converge_scf(self):
        """
        Submits :class:`aiida_fleur.workflows.scf.FleurScfWorkChain`.
        """
        inputs = {}
        if self.ctx.loop_count:
            inputs = self.get_inputs_scf()
        else:
            inputs = self.get_inputs_first_scf()
        res = self.submit(FleurScfWorkChain, **inputs)
        return ToContext(scf_res=res)

    def get_inputs_first_scf(self):
        """
        Initialize inputs for the first iteration.
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        if 'inpxml_changes' not in scf_wf_dict:
            scf_wf_dict['inpxml_changes'] = []

        scf_wf_dict['mode'] = 'force'

        if self.ctx.wf_dict['film_distance_relaxation']:
            scf_wf_dict['inpxml_changes'].append(
                ('set_atomgr_att', {'attributedict': {'force': [('relaxXYZ', 'FFT')]},
                                    'species': 'all'}))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        return input_scf

    def get_inputs_scf(self):
        """
        Initializes inputs for further iterations.
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))
        if 'structure' in input_scf:
            del input_scf.structure
            del input_scf.inpgen
            del input_scf.calc_parameters

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        scf_wf_dict['mode'] = 'force'
        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        scf_wc = self.ctx.scf_res
        last_calc = load_node(scf_wc.outputs.output_scf_wc_para.get_dict()['last_calc_uuid'])

        input_scf.remote_data = last_calc.outputs.remote_folder
        if self.ctx.new_fleurinp:
            input_scf.fleurinp = self.ctx.new_fleurinp

        return input_scf

    def check_failure(self):
        """
        Throws an exit code if scf failed
        """
        try:
            scf_wc = self.ctx.scf_res
        except AttributeError:
            message = 'ERROR: Something went wrong I do not have new atom positions calculation'
            self.control_end_wc(message)
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        if not scf_wc.is_finished_ok:
            exit_statuses = FleurScfWorkChain.get_exit_statuses(['ERROR_FLEUR_CALCULATION_FAILED'])
            if scf_wc.exit_status == exit_statuses[0]:
                fleur_calc = load_node(
                    scf_wc.outputs.output_scf_wc_para.get_dict()['last_calc_uuid'])
                if fleur_calc.exit_status == FleurCalc.get_exit_statuses(['ERROR_VACUUM_SPILL_RELAX'])[0]:
                    self.control_end_wc('Failed due to atom and vacuum overlap')
                    return self.exit_codes.ERROR_VACUUM_SPILL_RELAX
                elif fleur_calc.exit_status == FleurCalc.get_exit_statuses(['ERROR_MT_RADII_RELAX'])[0]:
                    self.control_end_wc('Failed due to MT overlap')
                    return self.exit_codes.ERROR_MT_RADII_RELAX
            return self.exit_codes.ERROR_SCF_FAILED

    def condition(self):
        """
        Checks if relaxation criteria is achieved.

        :return: True if structure is optimised and False otherwise
        """
        scf_wc = self.ctx.scf_res

        try:
            self.ctx.forces.append(scf_wc.outputs.output_scf_wc_para.dict.force_largest)
        except AttributeError:
            # message = 'ERROR: Did not manage to read the largest force'
            # self.control_end_wc(message)
            # return self.exit_codes.ERROR_RELAX_FAILED
            return False

        largest_now = abs(self.ctx.forces[-1])

        if largest_now < self.ctx.wf_dict['force_criterion']:
            self.report('Structure is converged to the largest force'
                        '{}'.format(self.ctx.forces[-1]))
            return False

        self.ctx.loop_count = self.ctx.loop_count + 1
        if self.ctx.loop_count == self.ctx.wf_dict['relax_iter']:
            self.ctx.reached_relax = False
            self.report('INFO: Reached optimization iteration number {}. Largest force is {}, '
                        'force criterion is {}'.format(self.ctx.loop_count + 1, largest_now,
                                                       self.ctx.wf_dict['force_criterion']))
            return False

        self.report('INFO: submit optimization iteration number {}. Largest force is {}, '
                    'force criterion is {}'.format(self.ctx.loop_count + 1, largest_now,
                                                   self.ctx.wf_dict['force_criterion']))

        return True

    def generate_new_fleurinp(self):
        """
        This function fetches relax.xml from the previous iteration and calls
        :meth:`~aiida_fleur.workflows.relax.FleurRelaxWorkChain.analyse_relax()`.
        New FleurinpData is stored in the context.
        """
        scf_wc = self.ctx.scf_res
        last_calc = load_node(scf_wc.outputs.output_scf_wc_para.get_dict()['last_calc_uuid'])
        try:
            relax_parsed = last_calc.outputs.relax_parameters
        except NotExistent:
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        new_fleurinp = self.analyse_relax(relax_parsed)

        self.ctx.new_fleurinp = new_fleurinp

    @staticmethod
    def analyse_relax(relax_dict):
        """
        This function generates a new fleurinp analysing parsed relax.xml from the previous
        calculation.

        **NOT IMPLEMENTED YET**

        :param relax_dict: parsed relax.xml from the previous calculation
        :return new_fleurinp: new FleurinpData object that will be used for next relax iteration
        """
        # TODO: implement this function, now always use relax.xml gemerated in FLEUR
        if False:
            return 1

        return None

    def get_results(self):
        """
        Generates results of the workchain.
        Creates a new structure data node which is an
        optimized structure.
        """
        try:
            relax_out = self.ctx.scf_res.outputs.last_fleur_calc_output
        except NotExistent:
            return self.exit_codes.ERROR_NO_SCF_OUTPUT

        relax_out = relax_out.get_dict()

        try:
            cell = relax_out['relax_brav_vectors']
            atom_positions = relax_out['relax_atom_positions']
            film = relax_out['film']
        except KeyError:
            return self.exit_codes.ERROR_NO_RELAX_OUTPUT

        self.ctx.final_cell = cell
        self.ctx.final_atom_positions = atom_positions

        if film == 'True':
            self.ctx.pbc = (True, True, False)
        else:
            self.ctx.pbc = (True, True, True)

    def return_results(self):
        """
        This function stores results of the workchain into the output nodes.
        """

        out = {'workflow_name': self.__class__.__name__,
               'workflow_version': self._workflowversion,
               'info': self.ctx.info,
               'warnings': self.ctx.warnings,
               'errors': self.ctx.errors,
               'force': self.ctx.forces,
               'force_iter_done': self.ctx.loop_count,
               'last_scf_wc_uuid': self.ctx.scf_res.uuid
               }

        if self.ctx.final_cell:
            structure = StructureData(cell=self.ctx.final_cell)
            bohr_a = 0.52917721092

            for atom in self.ctx.final_atom_positions:
                np_cell = np.array(self.ctx.final_cell)
                np_pos = np.array(atom[1:]) * bohr_a
                pos_abs = list(np.dot(np_cell, np_pos))
                if self.ctx.pbc == (True, True, True):
                    structure.append_atom(position=(pos_abs[0], pos_abs[1], pos_abs[2]),
                                          symbols=atom[0])
                else:  # assume z-direction is orthogonal to xy
                    structure.append_atom(position=(pos_abs[0], pos_abs[1], atom[3] * bohr_a),
                                          symbols=atom[0])

            structure.pbc = self.ctx.pbc
            structure = save_structure(structure)
            self.out('optimized_structure', structure)

        out = save_output_node(Dict(dict=out))
        self.out('out', out)
        if not self.ctx.reached_relax:
            return self.exit_codes.ERROR_DID_NOT_RELAX

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. It will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards.
        """
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def save_structure(structure):
    """
    Save a structure data node to provide correct provenance.
    """
    structure_return = structure.clone()
    return structure_return


@cf
def save_output_node(out):
    """
    Save the out dict in the db to provide correct provenance.
    """
    out_wc = out.clone()
    return out_wc



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the FleurBaseWorkChain.
FleurBaseWorkChain is a workchain that wraps the submission of
the FLEUR calculation. Inheritance from the BaseRestartWorkChain
allows to add scenarios to restart a calculation in an
automatic way if an expected failure occurred.
"""
from __future__ import absolute_import
from aiida_fleur.data.fleurinp import FleurinpData
import six

from aiida import orm
from aiida.common import AttributeDict
from aiida.engine import while_
from aiida.plugins import CalculationFactory, DataFactory
from aiida_fleur.common.workchain.base.restart import BaseRestartWorkChain
from aiida_fleur.tools.common_fleur_wf import optimize_calc_options
from aiida_fleur.common.workchain.utils import register_error_handler, ErrorHandlerReport

from aiida_fleur.calculation.fleur import FleurCalculation as FleurProcess


class FleurBaseWorkChain(BaseRestartWorkChain):
    """Workchain to run a FLEUR calculation with automated error handling and restarts"""
    _workflowversion = "0.1.1"

    _calculation_class = FleurProcess
    # _error_handler_entry_point = 'aiida_fleur.workflow_error_handlers.pw.base'

    @classmethod
    def define(cls, spec):
        super(FleurBaseWorkChain, cls).define(spec)
        spec.input('code', valid_type=orm.Code, help='The FLEUR code.')
        spec.input('parent_folder', valid_type=orm.RemoteData, required=False,
                   help='An optional working directory of a previously completed calculation to '
                   'restart from.')
        spec.input('settings', valid_type=orm.Dict, required=False,
                   help='Optional parameters to affect the way the calculation job and the parsing'
                   ' are performed.')
        spec.input('options', valid_type=orm.Dict,
                   help='Optional parameters to set up computational details.')
        spec.input('fleurinpdata', valid_type=FleurinpData,
                   help='Optional parameter set up a ready-to-use fleurinp.')
        spec.input('description', valid_type=six.string_types, required=False, non_db=True,
                   help='Calculation description.')
        spec.input('label', valid_type=six.string_types, required=False, non_db=True,
                   help='Calculation label.')

        spec.outline(
            cls.setup,
            cls.validate_inputs,
            while_(cls.should_run_calculation)(
                cls.run_calculation,
                cls.inspect_calculation,
            ),
            cls.results,
        )

        spec.output('output_parameters', valid_type=orm.Dict, required=False)
        spec.output('output_params_complex', valid_type=orm.Dict, required=False)
        spec.output('relax_parameters', valid_type=orm.Dict, required=False)
        spec.output('retrieved', valid_type=orm.FolderData, required=False)
        spec.output('remote_folder', valid_type=orm.RemoteData, required=False)
        spec.output('final_calc_uuid', valid_type=orm.Str, required=False)

        spec.exit_code(311, 'ERROR_VACUUM_SPILL_RELAX',
                       message='FLEUR calculation failed because an atom spilled to the'
                               'vacuum during relaxation')
        spec.exit_code(313, 'ERROR_MT_RADII_RELAX',
                       message='Overlapping MT-spheres during relaxation.')
        spec.exit_code(389, 'ERROR_MEMORY_ISSUE_NO_SOLUTION',
                       message="Computational resources are not optimal.")
        spec.exit_code(390, 'ERROR_NOT_OPTIMAL_RESOURCES',
                       message="Computational resources are not optimal.")
        spec.exit_code(399, 'ERROR_SOMETHING_WENT_WRONG',
                       message='FleurCalculation failed and FleurBaseWorkChain has no strategy '
                               'to resolve this')

    def validate_inputs(self):
        """
        Validate inputs that might depend on each other and cannot be validated by the spec.
        Also define dictionary `inputs` in the context, that will contain the inputs for the
        calculation that will be launched in the `run_calculation` step.
        """
        self.ctx.inputs = AttributeDict({
            'code': self.inputs.code,
            'fleurinpdata': self.inputs.fleurinpdata,
            'metadata': AttributeDict()
        })

        self.ctx.inputs.metadata.options = self.inputs.options.get_dict()

        if 'parent_folder' in self.inputs:
            self.ctx.inputs.parent_folder = self.inputs.parent_folder

        if 'description' in self.inputs:
            self.ctx.inputs.metadata.description = self.inputs.description
        else:
            self.ctx.inputs.metadata.description = ''
        if 'label' in self.inputs:
            self.ctx.inputs.metadata.label = self.inputs.label
        else:
            self.ctx.inputs.metadata.label = ''

        if 'settings' in self.inputs:
            self.ctx.inputs.settings = self.inputs.settings.get_dict()
        else:
            self.ctx.inputs.settings = {}

        resources_input = self.ctx.inputs.metadata.options['resources']
        try:
            self.ctx.num_machines = int(resources_input['num_machines'])
            self.ctx.num_mpiprocs_per_machine = int(resources_input['num_mpiprocs_per_machine'])
        except KeyError:
            self.ctx.can_be_optimised = False
            self.report('WARNING: Computation resources were not optimised.')
        else:
            try:
                self.ctx.num_cores_per_mpiproc = int(resources_input['num_cores_per_mpiproc'])
                self.ctx.use_omp = True
                self.ctx.suggest_mpi_omp_ratio = self.ctx.num_mpiprocs_per_machine / self.ctx.num_cores_per_mpiproc
            except KeyError:
                self.ctx.num_cores_per_mpiproc = 1
                self.ctx.use_omp = False
                self.ctx.suggest_mpi_omp_ratio = 1

            try:
                self.check_kpts()
                self.ctx.can_be_optimised = True
            except Warning:
                self.report('ERROR: Not optimal computational resources.')
                return self.exit_codes.ERROR_NOT_OPTIMAL_RESOURCES

    def check_kpts(self):
        """
        This routine checks if the total number of requested cpus
        is a factor of kpts and makes an optimisation.

        If suggested number of num_mpiprocs_per_machine is 60% smaller than
        requested, it throws an exit code and calculation stop withour submission.
        """
        fleurinp = self.ctx.inputs.fleurinpdata
        machines = self.ctx.num_machines
        mpi_proc = self.ctx.num_mpiprocs_per_machine
        omp_per_mpi = self.ctx.num_cores_per_mpiproc
        try:
            adv_nodes, adv_mpi_tasks, adv_omp_per_mpi, message = optimize_calc_options(
                machines, mpi_proc, omp_per_mpi, self.ctx.use_omp, self.ctx.suggest_mpi_omp_ratio, fleurinp)
        except ValueError:
            raise Warning('Not optimal computational resources, load less than 60%')

        self.report(message)

        self.ctx.inputs.metadata.options['resources']['num_machines'] = adv_nodes
        self.ctx.inputs.metadata.options['resources']['num_mpiprocs_per_machine'] = adv_mpi_tasks
        if self.ctx.use_omp:
            self.ctx.inputs.metadata.options['resources']['num_cores_per_mpiproc'] = adv_omp_per_mpi
            if self.ctx.inputs.metadata.options['environment_variables']:
                self.ctx.inputs.metadata.options['environment_variables']['OMP_NUM_THREADS'] = str(
                    adv_omp_per_mpi)
            else:
                self.ctx.inputs.metadata.options['environment_variables'] = {}
                self.ctx.inputs.metadata.options['environment_variables']['OMP_NUM_THREADS'] = str(
                    adv_omp_per_mpi)


@register_error_handler(FleurBaseWorkChain, 1)
def _handle_general_error(self, calculation):
    """
    Calculation failed for unknown reason.
    """
    if calculation.exit_status in FleurProcess.get_exit_statuses(['ERROR_FLEUR_CALC_FAILED',
                                                                  'ERROR_MT_RADII',
                                                                  'ERROR_NO_RETRIEVED_FOLDER',
                                                                  'ERROR_OPENING_OUTPUTS',
                                                                  'ERROR_NO_OUTXML',
                                                                  'ERROR_XMLOUT_PARSING_FAILED',
                                                                  'ERROR_RELAX_PARSING_FAILED']):
        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('Calculation failed for a reason that can not be resolved automatically')
        self.results()
        return ErrorHandlerReport(True, True, self.exit_codes.ERROR_SOMETHING_WENT_WRONG)
    else:
        raise ValueError('Calculation failed for unknown reason, please register the '
                         'corresponding exit code in this error handler')


@register_error_handler(FleurBaseWorkChain, 52)
def _handle_vacuum_spill_error(self, calculation):
    """
    Calculation failed for unknown reason.
    """
    if calculation.exit_status in FleurProcess.get_exit_statuses(['ERROR_VACUUM_SPILL_RELAX']):
        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('FLEUR calculation failed because an atom spilled to the vacuum during'
                    'relaxation. Can be fixed via RelaxBaseWorkChain.')
        self.results()
        return ErrorHandlerReport(True, True, self.exit_codes.ERROR_VACUUM_SPILL_RELAX)


@register_error_handler(FleurBaseWorkChain, 51)
def _handle_mt_relax_error(self, calculation):
    """
    Calculation failed for unknown reason.
    """
    if calculation.exit_status in FleurProcess.get_exit_statuses(['ERROR_MT_RADII_RELAX']):
        self.ctx.restart_calc = calculation
        self.ctx.is_finished = True
        self.report('FLEUR calculation failed due to MT overlap.'
                    ' Can be fixed via RelaxBaseWorkChain')
        self.results()
        return ErrorHandlerReport(True, True, self.exit_codes.ERROR_MT_RADII_RELAX)


@register_error_handler(FleurBaseWorkChain, 50)
def _handle_not_enough_memory(self, calculation):
    """
    Calculation failed due to lack of memory.
    Probably works for JURECA only, has to be tested for other systems.
    """

    if calculation.exit_status in FleurProcess.get_exit_statuses(['ERROR_NOT_ENOUGH_MEMORY']):
        if self.ctx.can_be_optimised:
            self.ctx.restart_calc = None
            self.ctx.is_finished = False
            self.report('Calculation failed due to lack of memory, I resubmit it with twice larger'
                        ' amount of computational nodes and smaller MPI/OMP ratio')
            self.ctx.num_machines = self.ctx.num_machines * 2
            self.ctx.suggest_mpi_omp_ratio = self.ctx.suggest_mpi_omp_ratio / 2
            self.check_kpts()

            if 'settings' not in self.ctx.inputs:
                self.ctx.inputs.settings = {}
            else:
                self.ctx.inputs.settings = self.inputs.settings.get_dict()
            self.ctx.inputs.settings.setdefault(
                'remove_from_remotecopy_list', []).append('mixing_history*')

            return ErrorHandlerReport(True, True)
        else:
            self.ctx.restart_calc = calculation
            self.ctx.is_finished = True
            self.report('FLEUR calculation failed due to MT overlap.'
                        ' Can be fixed via RelaxBaseWorkChain')
            self.results()
            return ErrorHandlerReport(True, True, self.exit_codes.ERROR_MEMORY_ISSUE_NO_SOLUTION)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# test the fleur specific graph gernation routine
from __future__ import absolute_import
import pytest

# These tests need dot/graphviz... which is not autoinstalled in the python env... so far
# Therefore I uncomment these tests for know, because they will fail on travis.
# TODO find a way (aiidas problem) to set up a clean environment
'''
# test draw_graph
@pytest.mark.usefixtures("aiida_env")
def test_draw_graph_if_produces_file():
    """
    does the individual fleur_draw_graph routine produce a file?
    """
    import os
    from aiida_fleur.tools.graph_fleur import draw_graph
    from aiida.orm import Node
    
    # TODO store a real graph and test if it is represented right...
    node = Node()
    outfile_expected = 'None.dot'
    exit_expected = 0
    exit_status, output_file_name = draw_graph(node)
    os.remove(output_file_name)

    assert exit_status == exit_expected
    assert output_file_name == outfile_expected

'''



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
Contains verdi commands for fleurinpdata
"""
from __future__ import absolute_import
import click

# this will get replaced by data_plug.group
#@data_plug.group('fleur.fleurinp')
@click.group()
def fleurinp():
    pass

@fleurinp.command()
def list_fleurinp():
    """
    list all Fleurinp data in the database and displays some information
    """
    click.echo('verdi data fleurinp list')
    #do a query and list all reuse AiiDA code

@fleurinp.command()
def show():
    """
    Shows some basic information about the fleurinp datastructure and dumbs the
    inp.xml
    """
    click.echo('verdi data fleurinp list')

@fleurinp.command()
def open_inp():
    """
    opens the inp.xml in some editor, readonly.
    inp.xml
    """
    click.echo('verdi data fleurinp list')

# this is a maybe
@fleurinp.command()
def get_structure():
    """
    Prints some basic information about the structure data and return a structure uuid/pk
    """
    click.echo('verdi data fleurinp list')


@fleurinp.command()
def get_kpoints():
    """
    Prints some basic information about the kpoints data and returns a kpoints uuid/pk
    """
    click.echo('verdi data fleurinp kpoints')
    
@fleurinp.command()
def get_parameters():
    """
    Prints some basic information about the parameter data and returns a 
    parameter data uuid/pk
    """
    click.echo('verdi data fleurinp parameter')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
DO NOT USE, this is crab so far. The development was stoped because this is done 
with AiiDA 'caching' now.

Here are methods to check the existence of something in the database
example if a (successful) SCF with the same inputs exists

Since cashing is not there for data yet, and below are some basic querries I 
leave the code here for now.
"""
from __future__ import absolute_import
from aiida.plugins import DataFactory
from aiida.orm import QueryBuilder
from aiida.engine.calculation.job import CalcJob
from aiida.orm import Node


'''
def check_existence_calc(input_nodes, successful=True):
    """
    This methods checks in the database waether a certain type of node with the given
    input nodes already exists. If yes it returns the output nodes of that node.

    param: input_nodes : List of input nodes

    returns output nodes
    """
    #TODO: some checks and inputnodes could be parsed in different formats
    inputnodesuuid = [node.uuid for node in input_nodes]

    qb=QueryBuilder()
    qb.append(
       JobCalculation, tag='calc', project='*',
       filters={'state' : {'==':'FINISHED'}})

    for idx, uuid in enumerate(inputnodesuuid):
        qb.append(Node, input_of='calc', filters={'uuid':uuid}, 
                  tag='input_{}'.format(idx))

    qb.order_by({JobCalculation:'ctime'})
    res = qb.all()
    if res:
        return res[-1][0].get_outputs()
    else:
        return None

def check_existence_wf(input_nodes, successful=True):
    """
    This methods checks in the database waether a certain type of node with the given
    input nodes already exists. If yes it returns the output nodes of that node.

    param: input_nodes : List of input nodes

    returns output nodes
    """
    #TODO: some checks and inputnodes could be parsed in different formats
    inputnodesuuid = [node.uuid for node in input_nodes]

    qb=QueryBuilder()
    qb.append(
       JobCalculation, tag='calc', project='*',
       filters={'state' : {'==':'FINISHED'}})

    for idx, uuid in enumerate(inputnodesuuid):
        qb.append(Node, input_of='calc', filters={'uuid':uuid}, tag='input_{}'.format(idx))

    qb.order_by({JobCalculation:'ctime'})
    res = qb.all()
    if res:
        return res[-1][0].get_outputs()
    else:
        return None

'''
'''
def intersectlist(l1, l2):
    common = []
    for element in l1:
        if element in l2:
            common.append(element)
    return common

def check_existence_calc(input_nodes, successful=True):
    """
    This methods checks in the database waether a certain type of node with the given
    input nodes already exists. If yes it returns the output nodes of that node.

    param: input_nodes : List of input nodes

    returns output nodes
    """
    inputnodesuuid = [node.uuid for node in input_nodes]
    overall_results = []

    for node in inputnodesuuid:
        suc = successful
        qb=QueryBuilder()
        qb.append(Node,
            filters={
                'uuid' : {'==': node},
            },
            tag='input')
        if suc:
            qb.append(
                JobCalculation,
                filters={
                    'state' : {'==':'FINISHED'}
                },
                output_of='input')
        else:
            qb.append(
                JobCalculation,
                output_of='input',
                project=['uuid'])
        res = qb.all()
        if res: # if there is no node with such an input return
            resnodesuuid = [node[0].uuid for node in res] # needed for common list parts
            overall_results.append(resnodesuuid)
        else:
            return None

    intersect = overall_results[0]
    if len(overall_results) > 1:
        for res in overall_results[1:]:
            intersect = intersectlist(intersect, res)
    qb1=QueryBuilder()
    qb1.append(
        JobCalculation,
        filters={
            'uuid' : {'in': intersect}
            })
    res = qb1.all()
    # we
    return res[0][0].outputs()
'''
'''
def check_existence(target_nodetype, input_nodes, successful=False):
    """
    This methods checks in the database waether a certain type of node with the given
    input nodes already exists. If yes it returns the output nodes of that node.

    param: target_nodetype
    param: input_nodes : List of input nodes

    returns output nodes

    Hints; successful is only for calculations types
    """
    inputnodesuuid = [node.uuid for node in input_nodes]
    qb=QueryBuilder()
    qb.append(Node,
        filters={
            'uuid' : {'in': inputnodesuuid},
        },
        tag='input')
    if successful:
        qb.append(
            target_nodetype,
            filters={
                'state' : {'==':'FINISHED'}
            },
            output_of='input')
    else:
        qb.append(
            target_nodetype,
            output_of='input')
    res = qb.all()
    print(len(res))
    if res:
        return res[0][0].ouputs()
    else:
        return None

def check_existence_calc(input_nodes, successful=True):
    """
    This methods checks in the database waether a certain type of node with the given
    input nodes already exists. If yes it returns the output nodes of that node.

    param: input_nodes : List of input nodes

    returns output nodes
    """
    inputnodesuuid = [node.uuid for node in input_nodes]
    qb=QueryBuilder()
    qb.append(Node,
        filters={
            'uuid' : {'in': inputnodesuuid},
        },
        tag='input')
    if successful:
        qb.append(
            JobCalculation,
            filters={
                'state' : {'==':'FINISHED'}
            },
            output_of='input')
    else:
        qb.append(
            JobCalculation,
            output_of='input')

    res = qb.all()
    print(len(res))
    if res:
        return res[0][0].ouputs()
    else:
        return None

def check_existence_wf(target_nodetype, input_nodes, successful=True):
    """
    This methods checks in the database waether a certain type of node with the given
    input nodes already exists. If yes it returns the output nodes of that node.

    param: input_nodes : List of input nodes

    returns output nodes
    """
    inputnodesuuid = [node.uuid for node in input_nodes]
    qb=QueryBuilder()
    qb.append(Node,
        filters={
            'uuid' : {'in': inputnodesuuid},
        },
        tag='input')
    if successful:
        qb.append(
            target_nodetype,
            filters={
                'state' : {'==':'FINISHED'}
            },
            output_of='input')
    else:
        qb.append(
            target_nodetype,
            output_of='input')
    res = qb.all()
    print(len(res))
    if res:
        return res[0][0].ouputs()
    else:
        return None
'''



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

from __future__ import absolute_import
import os
import pytest

# Collect the input files
file_path1 = '../files/inpxml/'
file_path2 = '../files/nonvalid_inpxml'

inpxmlfilefolder = os.path.dirname(os.path.abspath(__file__))
inpxmlfilefolder_valid = os.path.abspath(os.path.join(inpxmlfilefolder, file_path1))
inpxmlfilefolder_non_valid = os.path.abspath(os.path.join(inpxmlfilefolder, file_path2))

inpxmlfilelist = []
for subdir, dirs, files in os.walk(inpxmlfilefolder_valid):
    for file in files:
        if file.endswith('.xml'):
            inpxmlfilelist.append(os.path.join(subdir, file))

inpxmlfilelist2 = []
for subdir, dirs, files in os.walk(inpxmlfilefolder_non_valid):
    for file in files:
        if file.endswith('.xml'):
            inpxmlfilelist2.append(os.path.join(subdir, file))


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
def test_fleurinp_valid_inpxml(create_fleurinp, inpxmlfilepath):
    """
    test if valid inp.xml files are recognized by fleur
    """
    fleurinp_tmp = create_fleurinp(inpxmlfilepath)

    assert fleurinp_tmp._has_schema
    assert fleurinp_tmp._schema_file_path is not None
    assert fleurinp_tmp.inp_dict != {}
    assert fleurinp_tmp._validate() is None  # if fails, _validate throws an error


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist2)
def test_fleurinp_non_valid_inpxml(create_fleurinp, inpxmlfilepath):
    """
    Test what happens if inp.xml file is not valid
    either an InputValidationError should be thrown by fleurinpData or an
    XMLSyntaxError error if the inp.xml does not correspond to the xml schema.
    """
    from aiida.common.exceptions import InputValidationError
    from lxml import etree

    with pytest.raises((InputValidationError, etree.XMLSyntaxError)):
        create_fleurinp(inpxmlfilepath)


# test kpoints and structure and parameter data extraction

@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
def test_fleurinp_kpointsdata_extraction(create_fleurinp, inpxmlfilepath):
    """
    Extract a kpointsData from the fleurinp data, i.e inp.xml and check if
    the resulting node is a valid kpoints data
    """
    from aiida.orm import KpointsData

    fleurinp_tmp = create_fleurinp(inpxmlfilepath)
    kptsd = fleurinp_tmp.get_kpointsdata_ncf()

    if kptsd is not None:
        assert isinstance(kptsd, KpointsData)
    else:
        pass
        # What todo here, may test inpxml are with latnam definded, which does not work here.
        # or without a kpoint list. Therefore this test might let two much through


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
def test_fleurinp_parameterdata_extraction(create_fleurinp, inpxmlfilepath):
    """
    Extract a ParameterData from the fleurinp data, i.e inp.xml and check if
    the resulting node is a valid ParameterData. ggf if it can be used by inpgen
    """
    from aiida.orm import Dict

    fleurinp_tmp = create_fleurinp(inpxmlfilepath)
    param = fleurinp_tmp.get_parameterdata_ncf()

    assert isinstance(param, Dict)

    # ToDo check if it is also right for inpgen...


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
def test_fleurinp_structuredata_extraction(create_fleurinp, inpxmlfilepath):
    """
    Extract a ParameterData from the fleurinp data, i.e inp.xml and check if
    the resulting node is a valid ParameterData.
    """
    from aiida.orm import StructureData

    fleurinp_tmp = create_fleurinp(inpxmlfilepath)
    struc = fleurinp_tmp.get_structuredata_ncf()

    if struc is not None:
        assert isinstance(struc, StructureData)
    else:
        pass
        # What todo here, may test inpxml are with latnam definded,
        # which does not work here.
        # But if something else fails also None return. T
        # Therefore this test might let two much through


# Input Modification tests
@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
def test_fleurinp_single_value_modification(create_fleurinp, inpxmlfilepath):
    """
    set kmax, itmax, minDistance in inp.xml input file of fleurinpdata to
    10.2, 99, 0.000001, then check if it everything set
    """
    from aiida_fleur.data.fleurinpmodifier import FleurinpModifier

    fleurinp_tmp = create_fleurinp(inpxmlfilepath)
    fleurinpmode = FleurinpModifier(fleurinp_tmp)

    fleurinpmode.set_inpchanges({'itmax': 99, 'minDistance': 0.01, 'Kmax': 10.2})

    fleurinpmode.show(display=False, validate=True)
    out = fleurinpmode.freeze()

    assert isinstance(out, type(fleurinp_tmp))
    # TODO check if set right


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
def test_fleurinp_first_species_modification(create_fleurinp, inpxmlfilepath):
    """
    Decrease the rmt of the first species by 10%, check if rmt was set
    """
    pass
